!=================================================================
!
! NAME: 
!
!    Output_Values
!
! FUNCTION:
!
!    Outputs data from run
!
! TYPE/LANGUAGE: 
!
!    F90 module
!
! DESCRIPTION:  
!
!    Manages outputing data to files.  Also includes code to generate 
!    a non-opened file units
!
! ROUTINES TO BE USED EXTERNALLY:
!
!    Set_Output_File        - Opens output file
!    Close_Output_File      - Closes output file
!    Write_Output_File_Orig - Writes ASCII files (DW estimate only) in same 
!                             format as original code from Gary Wick
!    Write_Profiles_ASCII   - Write profile data to ASCII output
!    Get_File_Unit          - Gets new file unit
!
! ROUTINES USED INTERNALLY:
!
!    Write_Profile_ASCII    - Output profiles to ASCII file
!
! HISTORY:
! 
!   Original version by Jonathan P.D. Mittaz on 9/10/2013
!
!=================================================================
MODULE Output_Values

  USE Kinds
  USE Types
  USE ErrorHandler

  IMPLICIT NONE

  !
  ! Global variables
  !
  ! Output file unit only used internally to this module
  !
  INTEGER(Int4) :: Output_Unit
  INTEGER(Int4) :: Output_Profile_Unit

  PRIVATE
  PUBLIC :: Set_Output_File
  PUBLIC :: Close_Output_File
  PUBLIC :: Write_Output_File_Orig
  PUBLIC :: Get_File_Unit

CONTAINS

  !=================================================================
  !
  ! NAME:
  !
  !      Set_Output_File
  !
  ! FUNCTION:
  !
  !      Opens output file
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Opens output file and stores the file unit
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      Get_File_Unit
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Set_Output_File( params, profile_ascii)
  !
  ! INPUTS:
  ! 
  !      params        - Parameter structure
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !
  SUBROUTINE Set_Output_File( params )

    TYPE(Parameter_Type), INTENT(IN) :: params

    ! Local variables
    INTEGER(Int4) :: IOS
    
    IF( params%output_ascii_profiles )THEN
       Output_Profile_Unit = Get_File_Unit()
          
       OPEN(UNIT=Output_Profile_Unit,FILE=params%output_file_profile,&
            STATUS='UNKNOWN',IOSTAT=IOS)
       IF( 0 .ne. IOS )THEN
          CALL out_ERROR('Cannot open Profile Output file','Set_Output_File',&
               'Output_Values.f90')
       ENDIF
    ENDIF
    
    IF( params%Output_File_NetCDF_Format )THEN
       CALL out_ERROR('NetCDF output format not currently supported',&
            'Set_Output_File','Output_Values.f90')
    ELSE
       Output_Unit = Get_File_Unit()
       
       OPEN(UNIT=Output_Unit,FILE=params%output_file,STATUS='UNKNOWN',&
            IOSTAT=IOS)
       IF( 0 .ne. IOS )THEN
          CALL out_ERROR('Cannot open Output file','Set_Output_File',&
               'Output_Values.f90')
       ENDIF
    ENDIF

  END SUBROUTINE Set_Output_File

  !=================================================================
  !
  ! NAME:
  !
  !      Close_Output_File
  !
  ! FUNCTION:
  !
  !      Closes output file
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Closes output file 
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Close_Output_File( params )
  !
  ! INPUTS:
  ! 
  !      params    - Parameter structure
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !
  SUBROUTINE Close_Output_File( params )

    TYPE(Parameter_Type), INTENT(IN) :: params

    !
    ! Close profile file if opened
    IF( params%output_ascii_profiles )THEN
       CLOSE(Output_Profile_Unit)
    ENDIF
    
    IF( params%Output_File_NetCDF_Format )THEN
       CALL out_ERROR('NetCDF output format not currently supported',&
            'Close_Output_File','Output_Values.f90')
    ELSE
       CLOSE(Output_Unit)
    ENDIF

  END SUBROUTINE Close_Output_File

  !=================================================================
  !
  ! NAME:
  !
  !      Write_Output_File_Orig
  !
  ! FUNCTION:
  !
  !      Writes diurnal warming estimate to ASCII file
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Writes out dirunal warming estimate (T(1)-T(5m)) to ASCII format
  !      with the same format as the original code
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Write_Output_File_Orig( iTime, params, surface, profile, 
  !                                   Header )
  !
  ! INPUTS:
  ! 
  !      iTime     - Time step from time loop in main code
  !      params    - Input parameters
  !      surface   - Surface and time data
  !      profile   - Profile data
  !      Header    - Logical to enabel writing header information
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !
  SUBROUTINE Write_Output_File_Orig( iTime, params, surface, &
       profile, Header )

    INTEGER(Int4), INTENT(IN) :: iTime
    TYPE(parameter_type), INTENT(IN) :: params
    TYPE(Surface_Type), INTENT(IN) :: surface
    TYPE(Profile_Type), INTENT(IN) :: profile
    LOGICAL, INTENT(IN) :: Header

    IF( params%Output_File_NetCDF_Format )THEN
       CALL out_ERROR('NetCDF output format not currently supported',&
            'Set_Output_File','Output_Values.f90')
    ELSE
       IF( Header )THEN
          WRITE(Output_Unit,'(I8,1x,3(F10.4,1x),2(I4,1x))')&
               surface%date,surface%latitude,surface%longitude,&
               surface%sstinit,surface%i_out,surface%j_out
       ELSE
          IF( iTime .ne. 1 .and. &
               MOD((iTime-1),params%print_freq_loop) .eq. 0 )THEN
             !
             ! test_output is Gary Wicks reduced output format
             !
             IF( params%test_output )THEN
                write(Output_Unit,'(F13.6,1x,7(F10.4,1x))')surface%time, &
                     profile%temp(1), surface%tskin, profile%temp(70), &
                     profile%temp(30), profile%temp(115), surface%qsol, &
                     surface%wind_speed
             ELSE
                !
                ! Just output the DW estimate
                !
                IF( params%use_skin_temperature )THEN
                   !
                   ! Use Tskin
                   !
                   WRITE(Output_Unit,'(F13.6,1x,F10.4)')surface%time,&
                        surface%tskin-profile%temp(profile%pos_reference)
                ELSE
                   !
                   ! Use Tsubskin
                   !
                   WRITE(Output_Unit,'(F13.6,1x,F10.4)')surface%time,&
                        profile%temp(1)-profile%temp(profile%pos_reference)
                ENDIF
             ENDIF
          ENDIF
       ENDIF
    ENDIF

    !
    ! If writing out ASCII profiles
    IF( params%output_ascii_profiles .and. .not.Header )THEN
       CALL Write_Profile_ASCII(surface, profile)
    ENDIF

  END SUBROUTINE Write_Output_File_Orig

  !=================================================================
  !
  ! NAME:
  !
  !      Write_Profile_ASCII
  !
  ! FUNCTION:
  !
  !      Writes profile data to ASCII file
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Writes out profiles to an ASCII file
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Write_Profile_ASCII( surface, profile )
  !
  ! INPUTS:
  ! 
  !      surface   - Surface and time data
  !      profile   - Profile data
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !
  SUBROUTINE Write_Profile_ASCII( surface, profile )

    TYPE(Surface_Type), INTENT(IN) :: surface
    TYPE(Profile_Type), INTENT(IN) :: profile

    ! Local variables
    
    WRITE(Output_Profile_Unit,*)surface%time
    WRITE(Output_Profile_Unit,*)profile%nlevels
    WRITE(Output_Profile_Unit,*)surface%wind_speed,surface%qsol
    WRITE(Output_Profile_Unit,*)profile%z*profile%bottom
    WRITE(Output_Profile_Unit,*)profile%swrad
    WRITE(Output_Profile_Unit,*)profile%dm_temp
    WRITE(Output_Profile_Unit,*)profile%temp
    WRITE(Output_Profile_Unit,*)profile%salinity
    WRITE(Output_Profile_Unit,*)profile%Velocity_U
    WRITE(Output_Profile_Unit,*)profile%Velocity_V
    WRITE(Output_Profile_Unit,*)profile%Velocity_U_Stokes
    WRITE(Output_Profile_Unit,*)profile%Velocity_V_Stokes
    WRITE(Output_Profile_Unit,*)profile%Q2
    WRITE(Output_Profile_Unit,*)profile%Q2L
    WRITE(Output_Profile_Unit,*)profile%rho
    WRITE(Output_Profile_Unit,*)profile%richardson_no
    WRITE(Output_Profile_Unit,*)profile%shear_prod
    WRITE(Output_Profile_Unit,*)profile%buoyancy_prod
    WRITE(Output_Profile_Unit,*)profile%shear_prod_stokes
    WRITE(Output_Profile_Unit,*)profile%prod
    WRITE(Output_Profile_Unit,*)profile%KQ
    WRITE(Output_Profile_Unit,*)profile%KN
    WRITE(Output_Profile_Unit,*)profile%KM
    WRITE(Output_Profile_Unit,*)profile%KH

  END SUBROUTINE Write_Profile_ASCII

  !=================================================================
  !
  ! NAME:
  !
  !      Get_File_Unit
  !
  ! FUNCTION:
  !
  !      Gets an available file unit
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Get the next available file unit.  Ensures file units are available
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      unit = Get_File_Unit()
  !
  ! INPUTS:
  !
  ! OUTPUTS:
  !
  ! RETURNS:
  !
  !       new file unit
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  INTEGER(Int4) FUNCTION Get_File_Unit()

    ! Local variables
    INTEGER(Int4) :: I
    LOGICAL :: open_logical

    Get_File_Unit = -1
    unit_loop: DO I=20,200
       INQUIRE(Unit=I,OPENED=open_logical)
       IF( .not. open_logical )THEN
          Get_File_Unit = I
          EXIT unit_loop
       ENDIF
    END DO unit_loop

    IF( -1 .eq. Get_File_Unit )THEN
       CALL out_ERROR('Ran out of file units','Get_File_Unit','get_data.f90')
    ENDIF

  END FUNCTION Get_File_Unit

END MODULE Output_Values
