!=================================================================
!
! NAME: 
!
!    Get_Data
!
! FUNCTION:
!
!    Reads in model and assiated data
!
! TYPE/LANGUAGE: 
!
!    F90 module
!
! DESCRIPTION:  
!
!    Reads in initial SST and model data and time interpolates model data 
!    to time step
!
! ROUTINES TO BE USED EXTERNALLY:
!
!    Access_Data_Files    - Open model data files and read in parameters
!                           from parameter file as well as the turbulent
!                           constants
!    Close_Data_Files     - Close all open model files
!    Read_in_grid         - Reads in grid (sst init data)
!    Get_Fluxes           - Reads in model flux data including wind model
!
! ROUTINES USED INTERNALLY:
!
!    Read_Turbulence_File - Reads in turbulence data
!    Modify_Fluxes        - Modifies the flux data on the basis of any 
!                           diurnal warming
!    humidity             - Get the humidity at the surface
!    Get_Stokes           - Get stokes drift information based on wind model
!
! HISTORY:
! 
!   Original version by Jonathan P.D. Mittaz on 9/10/2013
!
!=================================================================
MODULE Get_Data

  USE Kinds
  USE Types
  USE Constants
  USE Solar_Routines
  USE ErrorHandler
  USE get_parameters
  USE profile
  USE Output_Values

  IMPLICIT NONE

  !
  ! Global variables for this module
  !
  INTEGER(Int4) :: Unit_Grid
  INTEGER(Int4) :: Unit_Fluxes

  PRIVATE
  PUBLIC :: Access_Data_Files
  PUBLIC :: Close_Data_Files
  PUBLIC :: Read_in_grid
  PUBLIC :: Get_Fluxes

CONTAINS
  
  !=================================================================
  !
  ! NAME:
  !
  !      Access_Data_Files
  !
  ! FUNCTION:
  !
  !      Gets input files from the command line and reads them into the
  !      relevant structures.  
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Gets the required input files from the command line and reads the
  !      the paramater file as well as the turbulence parameters.  Also 
  !      opens the data streams for the input fluxes.
  ! 
  !      This version reads the original versions of the file generated by
  !      Gary Wicks IDL code, not the new format files.  
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      IARGC
  !      GETARG
  !      Read_Parameter_File
  !      Read_Turbulence_File
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Access_Data_Files( params, turb_cnts )
  !
  ! INPUTS:
  ! 
  ! OUTPUTS:
  !
  !      params    - Parameters needed to run the code
  !      turb_cnts - Turbulence constants
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  SUBROUTINE Access_Data_Files( params, turb_cnts )

    TYPE(parameter_type), INTENT(OUT) :: params
    TYPE(Turb_Cnts_Type), INTENT(OUT) :: turb_cnts

    ! Local variables
    INTEGER(Int4) :: nargs
    INTEGER(Int4) :: IOS
    CHARACTER(LEN=256) :: File_parameter
    CHARACTER(LEN=256) :: Model_Dir
    CHARACTER(LEN=256) :: DateString

    !
    ! Get filenames from command file
    !
    ! Check number of arguments
    !
    nargs = IARGC()
    IF( 3 .ne. nargs )THEN
       WRITE(*,'('' USAGE:'')')
       WRITE(*,'(''        kantha_clayson_dwmodel.exe param_file model_dir date_string'')')
       WRITE(*,'(''            param_file      - Input parameter file'')')
       WRITE(*,'(''            model_dir       - Directory for model input/ouput'')')
       WRITE(*,'(''            date_string     - Date string (YYYYMMDD)'')')
       STOP 1
    ENDIF

    !
    ! Get parameter filename
    !
    CALL GETARG(1,File_parameter)

    !
    ! Get directory where model data will be stored
    !
    CALL GETARG(2,Model_Dir)

    ! 
    ! Get Date string for file input/output
    !
    CALL GETARG(3,DateString)

    !
    ! Read in parameter file
    !
    CALL Read_Parameter_File(File_parameter,params,Model_Dir,DateString)

    !
    ! Read in turbulence parameters file
    !
    CALL Read_Turbulence_File(params%File_turb_cnts,params,turb_cnts)

    !
    ! Open Model data
    !
    IF( params%use_ascii_input )THEN
       !
       ! Open grid file
       !
       Unit_Grid = Get_File_Unit()
       OPEN(Unit=Unit_Grid,FILE=params%File_grid,STATUS='OLD',IOSTAT=IOS)
       IF( 0 .ne. IOS )THEN
          CALL out_ERROR('Cannot open GRID file','Access_Data_Files',&
               'get_data.f90')
       ENDIF
       
       !
       ! Open Fluxes file
       !
       Unit_fluxes = Get_File_Unit()
       OPEN(Unit=Unit_fluxes,FILE=params%File_fluxes,STATUS='OLD',IOSTAT=IOS)
       IF( 0 .ne. IOS )THEN
          CALL out_ERROR('Cannot open FLUXES file','Access_Data_Files',&
               'get_data.f90')
       ENDIF

       !
       ! Open output file
       !
       CALL Set_Output_File(params)
    
    ELSE

       CALL out_ERROR('Only OLD ASCII file format supported',&
            'Access_Data_Files','get_data.f90')

    ENDIF

  END SUBROUTINE Access_Data_Files

  !=================================================================
  !
  ! NAME:
  !
  !      Read_Turblence_File
  !
  ! FUNCTION:
  !
  !      Reads in the turbulence constants from a file
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Reads in the turbulence constants from a file
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Read_Turbulence_File( filename, turb_cnts )
  !
  ! INPUTS:
  !
  !      filename  - input filename
  !
  ! OUTPUTS:
  !
  !      turb_cnts - Turbulence constants
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  SUBROUTINE Read_Turbulence_File( filename, params, turb_cnts )

    CHARACTER(LEN=*), INTENT(IN) :: filename
    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(turb_cnts_type), INTENT(OUT) :: turb_cnts

    ! Local variables
    INTEGER :: IOS
    INTEGER :: Unit

    Unit = Get_File_Unit()
    OPEN(UNIT=Unit,FILE=filename,STATUS='OLD',IOSTAT=IOS)
    IF( 0 .ne. IOS )THEN
       CALL out_ERROR('Cannot open Turbulence File','Read_Turbulence_File',&
            'get_data.f90')
    ENDIF

    READ(Unit,*,IOSTAT=IOS)turb_cnts%A1,turb_cnts%B1,turb_cnts%A2,&
         turb_cnts%B2,turb_cnts%C1,turb_cnts%C2,turb_cnts%C3
    IF( 0 .ne. IOS )THEN
       CALL out_ERROR('Cannot read Turbulence File','Read_Turbulence_File',&
            'get_data.f90')
    ENDIF

    READ(Unit,*,IOSTAT=IOS)turb_cnts%E1,turb_cnts%E2,turb_cnts%E3,&
         turb_cnts%E4,turb_cnts%E5,turb_cnts%E6
    IF( 0 .ne. IOS )THEN
       CALL out_ERROR('Cannot read Turbulence File (line 2)',&
            'Read_Turbulence_File','get_data.f90')
    ENDIF

    READ(Unit,*,IOSTAT=IOS)turb_cnts%slsq
    IF( 0 .ne. IOS )THEN
       CALL out_ERROR('Cannot read Turbulence File (Line 3)',&
            'Read_Turbulence_File','get_data.f90')
    ENDIF
    
    CLOSE(Unit)

    !
    ! Calculate extra constants
    !
    turb_cnts%CF1 = turb_cnts%A2 * &
         (1.0_RealType-6.0_RealType*turb_cnts%A1/turb_cnts%B1)
    turb_cnts%CF2 = 3.0_RealType*turb_cnts%A2*&
         (6.0_RealType*turb_cnts%A1+turb_cnts%B2*(1.0_RealType-turb_cnts%C3))
    turb_cnts%CF3 = turb_cnts%A1*&
         (1.0_RealType-3.0_RealType*turb_cnts%C1 - &
         6.0_RealType*turb_cnts%A1/turb_cnts%B1)
    turb_cnts%CF4=9.0_RealType*turb_cnts%A1 * &
         (2.0_RealType*turb_cnts%A1+turb_cnts%A2*&
         (1.0_RealType-turb_cnts%C2))
    turb_cnts%CF5=9.0_RealType*turb_cnts%A1*turb_cnts%A2

    turb_cnts%GH_min = -0.25_RealType
    turb_cnts%GH_max = 0.98_RealType/(turb_cnts%A2*(turb_cnts%B1+&
         12.0_RealType*turb_cnts%A1 + &
         3.0_RealType*turb_cnts%B2*(1.0_RealType-turb_cnts%C3)))

    turb_cnts%total_KE_Const = 6.5073672_RealType * &
         params%shear_enhancement_factor
    
  END SUBROUTINE Read_Turbulence_File

  !=================================================================
  !
  ! NAME:
  !
  !      Close_Data_Files
  !
  ! FUNCTION:
  !
  !      Closes the model flux data files
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Closes the model flux data files.
  !
  !      Note this is for the old style (Gary Wick IDL generated) data
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Close_Data_Files(params)
  !
  ! INPUTS:
  !
  !      params   - parameter structure
  !
  ! OUTPUTS:
  !
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  SUBROUTINE Close_Data_Files(params)

    TYPE(Parameter_Type), INTENT(IN) :: params

    IF( params%use_ascii_input )THEN
       CLOSE(Unit_Grid)
       CLOSE(Unit_Fluxes)
    ELSE
       CALL out_ERROR('Only OLD ASCII model format supported',&
            'Close_Data_Files','get_data.f90')
    ENDIF

  END SUBROUTINE Close_Data_Files

  !=================================================================
  !
  ! NAME:
  !
  !      Read_in_Grid
  !
  ! FUNCTION:
  !
  !      Reads in the simple (lat/lon/SST) information for the flux grid
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Reads on the simple information for the flux information
  !
  !      Note this is for the old style (Gary Wick IDL generated) data
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Read_in_Grid(params, surface)
  !
  ! INPUTS:
  !
  !      params    - input parameters
  !
  ! OUTPUTS:
  !
  !      surface   - surface structure
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  SUBROUTINE Read_in_Grid(params,surface)

    TYPE(Parameter_Type), INTENT(INOUT) :: params
    TYPE(Surface_Type), INTENT(INOUT) :: surface

    ! Local variables
    INTEGER(Int4) :: IOS
    INTEGER(Int4) :: itemp

    IF( params%use_ascii_input )THEN

       READ(Unit_grid,*,IOSTAT=IOS)surface%date,surface%latitude,&
            surface%longitude,surface%sstinit,surface%i_out,surface%j_out
       IF( 0 .ne. IOS )THEN
          CALL out_ERROR('Error reading/End of file for GRID','Read_in_Grid',&
               'get_data.f90')
       ENDIF

       !
       ! Make i_out and j_out start from an index of 1
       !
       surface%i_out = surface%i_out + 1
       surface%j_out = surface%j_out + 1

       !
       ! If date not obtained, get it 
       !
       
       surface%Year = surface%date/10000
       IF( .not.params%date_there )THEN
          ! 
          ! Get year, month, day from date
          !
          params%Year = surface%Year
          itemp = surface%date-params%Year*10000
          params%Month = itemp/100
          params%Day = itemp - params%Month*100
          
          !
          ! Get DateString
          !
          WRITE(params%DateString,'(i4.4,i2.2,i2.2)')params%Year,&
               params%Month,params%Day
          params%date_there = .TRUE.
       ENDIF

    ELSE

       CALL out_ERROR('Only OLD binary file format supported',&
            'Read_in_Grid','get_data.f90')

    ENDIF

  END SUBROUTINE Read_in_Grid

  !=================================================================
  !
  ! NAME:
  !
  !      Get_Fluxes
  !
  ! FUNCTION:
  !
  !      Reads in model data and generates all fluxes
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Reads in binary model data and calculates solar fluxes etc.
  !
  !      Note this is for the old style (Gary Wick IDL generated) data
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Get_Fluxes(istep,params,profile,surface)
  !
  ! INPUTS:
  !
  !      istep   - where in time iteration - if 1 then start from beginning
  !
  ! OUTPUTS:
  !
  ! INPUTS/OUTPUTS:
  !
  !      params  - input parameters - use_langmuir can be modified
  !      profile - input depth profile
  !      surface  - structure contaings surface/initialization data
  ! 
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  SUBROUTINE Get_Fluxes( istep, params, profile, surface )

    INTEGER, INTENT(IN) :: istep
    TYPE(parameter_type), INTENT(INOUT) :: params
    TYPE(profile_type), INTENT(INOUT) :: profile
    TYPE(surface_type), INTENT(INOUT) :: surface

    ! Local variables
    INTEGER(Int4) :: IOS
    REAL(RealType) :: fraction
    REAL(RealType) :: sat_za
    REAL(RealType) :: tot_solar_clear_sky
    REAL(RealType) :: TauTot
    REAL(RealType) :: GustFactor
    REAL(RealType) :: Gust
    REAL(RealType) :: cli
    REAL(RealType) :: cli_dummy
    REAL(RealType) :: total_heat
    REAL(RealType) :: total_solar
    REAL(Dble) :: newtime

    !
    ! If first time - read in data
    !
    IF( 1 .eq. istep )THEN

       IF( params%use_ascii_input )THEN
          READ(Unit_Fluxes,*,IOSTAT=IOS)surface%time1,&
               surface%wind_stress_x1,surface%wind_stress_y1,&
               surface%downwelling_solar1,surface%heat_flux_longwave1,&
               surface%heat_flux_latent1,surface%heat_flux_sensible1,&
               surface%significant_wave_height1,surface%wave_period1,&
               surface%wave_dir1,surface%model_sst1,&
               surface%temp_atmos1,surface%specific_humidity_atmos1
          IF( 0 .ne. IOS )THEN
             CALL out_ERROR('End of file Fluxes file','Get_Fluxes',&
                  'get_data.f90')
          ENDIF

          READ(Unit_Fluxes,*,IOSTAT=IOS)surface%time2,&
               surface%wind_stress_x2,surface%wind_stress_y2,&
               surface%downwelling_solar2,surface%heat_flux_longwave2,&
               surface%heat_flux_latent2,surface%heat_flux_sensible2,&
               surface%significant_wave_height2,surface%wave_period2,&
               surface%wave_dir2,surface%model_sst2,&
               surface%temp_atmos2,surface%specific_humidity_atmos2
          IF( 0 .ne. IOS )THEN
             CALL out_ERROR('End of file Fluxes file','Get_Fluxes',&
                  'get_data.f90')
          ENDIF

          !
          ! Get time
          !
          newtime = surface%Time1 + (istep-0.5_Dble)*params%dt1 / &
               86400.0_Dble
          surface%start_time = surface%time1
          surface%time = surface%start_Time

          !
          ! For the minute have surface pressure as 1008.
          !
          surface%pressure1 = 1008.0_RealType
          surface%pressure2 = 1008.0_RealType
       
       ELSE

          CALL out_ERROR('Only OLD binary file format supported',&
               'Read_in_Grid','get_data.f90')
       
       ENDIF

       !
       ! Set time
       !
       surface%start_time = surface%time1
       surface%time = surface%start_Time

       !
       ! Define clear sky fraction for both times - if solar flux large enough
       ! Note that index = 0. for clear sky, 1. for cloudy
       !      
       IF( ABS(surface%downwelling_solar1) .ge. min_solar )THEN
          CALL Solar_Properties(surface%time1,surface%latitude,&
               surface%longitude,-surface%downwelling_solar1,sat_za,&
               surface%clear_sky_index1,tot_solar_clear_sky)
       ELSE
          surface%clear_sky_index1 = 1.0_RealType
       ENDIF

       IF( ABS(surface%downwelling_solar2) .ge. min_solar )THEN
          CALL Solar_Properties(surface%time2,surface%latitude,&
               surface%longitude,-surface%downwelling_solar2,sat_za,&
               surface%clear_sky_index2,tot_solar_clear_sky)
       ELSE
          surface%clear_sky_index2 = 1.0_RealType
       ENDIF

       surface%fluxes_init = .TRUE.
    ELSE

       IF( .not. surface%fluxes_init )THEN
          CALL out_ERROR('Fluxes not initialized','Get_Fluxes','get_data.f90')
       ENDIF

       ! 
       ! For all other cases - work out if have crossed to next NWP data
       !
       ! newtime is at 1/2 time step
       !
       newtime = surface%start_Time + (istep-0.5_Dble)*params%dt1 / &
            86400.0_Dble
       surface%time = surface%start_Time + (istep-1)*params%dt1 / &
            86400.0_Dble
       IF( newtime .gt. surface%time2 )THEN

          !
          ! Move time2 to time1 and read in new time2 values
          !
          surface%time1 = surface%time2
          surface%wind_stress_x1 = surface%wind_stress_x2
          surface%wind_stress_y1 = surface%wind_stress_y2
          surface%downwelling_solar1 = surface%downwelling_solar2
          surface%heat_flux_longwave1 = surface%heat_flux_longwave2
          surface%heat_flux_latent1 = surface%heat_flux_latent2
          surface%heat_flux_sensible1 = surface%heat_flux_sensible2
          surface%significant_wave_height1 = surface%significant_wave_height2
          surface%wave_period1 = surface%wave_period2
          surface%wave_dir1 = surface%wave_dir2
          surface%model_sst1 = surface%model_sst2
          surface%temp_atmos1 = surface%temp_atmos2
          surface%specific_humidity_atmos1 = surface%specific_humidity_atmos2
          surface%clear_sky_index1 = surface%clear_sky_index2 

          !
          ! Read in next time step
          !
          IF( params%use_ascii_input )THEN
             READ(Unit_Fluxes,*,IOSTAT=IOS)surface%time2,&
                  surface%wind_stress_x2,surface%wind_stress_y2,&
                  surface%downwelling_solar2,surface%heat_flux_longwave2,&
                  surface%heat_flux_latent2,surface%heat_flux_sensible2,&
                  surface%significant_wave_height2,surface%wave_period2,&
                  surface%wave_dir2,surface%model_sst2,&
                  surface%temp_atmos2,surface%specific_humidity_atmos2
             IF( 0 .ne. IOS )THEN
                CALL out_ERROR('End of file Fluxes file','Get_Fluxes',&
                     'get_data.f90')
             ENDIF
          ELSE

             CALL out_ERROR('Only OLD binary file format supported',&
                  'Read_in_Grid','get_data.f90')

          ENDIF

          !
          ! Get new clear sky index
          !          
          IF( ABS(surface%downwelling_solar2) .ge. min_solar )THEN
             CALL Solar_Properties(surface%time2,surface%latitude,&
                  surface%longitude,-surface%downwelling_solar2,sat_za,&
                  surface%clear_sky_index2,tot_solar_clear_sky)
          ELSE
             surface%clear_sky_index2 = 1.0_RealType
          ENDIF

       ENDIF

       IF( newtime .gt. surface%time2 )THEN
          CALL out_ERROR('time2 < current time','Get_Fluxes','get_data.f90')
       ENDIF

    ENDIF

    !
    ! Get time fraction between time1 to time2
    !
    fraction = REAL((newtime - surface%time1)/(surface%time2-surface%time1),&
         KIND=RealType)

    !
    ! Interpolate all but solar intensity to current time
    !
    surface%wind_stress_x = (1._RealType-fraction)*surface%wind_stress_x1 + &
         fraction*surface%wind_stress_x2
    surface%wind_stress_y = (1._RealType-fraction)*surface%wind_stress_y1 + &
         fraction*surface%wind_stress_y2
    surface%downwelling_solar = (1._RealType-fraction)*surface%downwelling_solar1 + &
         fraction*surface%downwelling_solar2
    surface%heat_flux_longwave = (1._RealType-fraction) * &
         surface%heat_flux_longwave1 + &
         fraction*surface%heat_flux_longwave2
    surface%heat_flux_latent = (1._RealType-fraction) * &
         surface%heat_flux_latent1 + &
         fraction*surface%heat_flux_latent2
    surface%heat_flux_sensible = (1._RealType-fraction) * &
         surface%heat_flux_sensible1 + &
         fraction*surface%heat_flux_sensible2
    surface%significant_wave_height = (1._RealType-fraction) * &
         surface%significant_wave_height1 + &
         fraction*surface%significant_wave_height2
    surface%wave_period = (1._RealType-fraction) * &
         surface%wave_period1 + fraction*surface%wave_period2
    surface%wave_dir = (1._RealType-fraction) * &
         surface%wave_dir1 + fraction*surface%wave_dir2
    surface%model_sst = (1._RealType-fraction) * &
         surface%model_sst1 + fraction*surface%model_sst2
    surface%temp_atmos = (1._RealType-fraction) * &
         surface%temp_atmos1 + fraction*surface%temp_atmos2
    surface%specific_humidity_atmos = (1._RealType-fraction) * &
         surface%specific_humidity_atmos1 + &
         fraction*surface%specific_humidity_atmos2
    surface%pressure = (1._RealType-fraction)*surface%pressure1 + &
         fraction*surface%pressure2

    surface%heat_flux_longwave = (1._RealType-fraction) * &
         surface%heat_flux_longwave1 + &
         fraction*surface%heat_flux_longwave2

    !
    ! If langmuir circulation/stokes drift requested, check to see if wave 
    ! information exists
    !
    IF( params%use_langmuir_param )THEN
       IF( surface%significant_wave_height1 .lt. 0._RealType .or. &
            surface%significant_wave_height2 .lt. 0._RealType  )THEN
          !
          ! No wave information - ignore langmuir circ/stokes drift for this 
          ! position
          !
          params%use_langmuir = .FALSE.
       ELSE
          !
          ! Wave information exists - do langmuir circ/stokes drift
          !
          params%use_langmuir = .TRUE.
       ENDIF
    ENDIF

    !
    ! breakwave turbulent scale
    !
    surface%breakwave_turb_scale = 1.6_RealType*surface%significant_wave_height
    
    !
    ! For solar, interpolation will be based on the clear sky index.
    ! Since the values are unreliable at nighttime, we use a different logic
    ! based on which of the bracketing observations are daytime.
    !
    
    !
    ! If nighttime for both
    !
    IF( ABS(surface%downwelling_solar1) .lt. min_solar .and. &
         ABS(surface%downwelling_solar2) .lt. min_solar )THEN
       !
       ! solar radiation == 0.0
       !
       surface%qsol = 0.0_RealType
    !
    ! If nighttime for second
    !
    ELSE IF( ABS(surface%downwelling_solar1) .ge. min_solar .and. &
         ABS(surface%downwelling_solar2) .lt. min_solar )THEN
       cli = surface%clear_sky_index1
       !
       ! Get clear sky solar
       !
       CALL Solar_Properties(newtime,surface%latitude,surface%longitude,&
            0.0_RealType,sat_za,cli_dummy,total_solar)
       surface%qsol = -((1.0_RealType-cli) * total_solar)
    !
    ! If nighttime for first
    !
    ELSE IF( ABS(surface%downwelling_solar1) .lt. min_solar .and. &
         ABS(surface%downwelling_solar2) .ge. min_solar )THEN
       cli = surface%clear_sky_index2
       !
       ! Get clear sky solar
       !
       CALL Solar_Properties(newtime,surface%latitude,surface%longitude,&
            0.0_RealType,sat_za,cli_dummy,total_solar)
       surface%qsol = -((1.0_RealType-cli) * total_solar)
    !
    ! If daytime for both
    !
    ELSE IF( ABS(surface%downwelling_solar1) .ge. min_solar .and. &
         ABS(surface%downwelling_solar2) .ge. min_solar )THEN
       cli = (1.-fraction)*surface%clear_sky_index1 + &
            fraction*surface%clear_sky_index2
       !
       ! Get clear sky solar
       !
       CALL Solar_Properties(newtime,surface%latitude,surface%longitude,&
            0.0_RealType,sat_za,cli_dummy,total_solar)
       surface%qsol = -((1.0_RealType-cli) * total_solar)
    ENDIF

    !
    ! If there was an error, set solar to 0.0
    !
    IF( surface%qsol .lt. -solar_constant )THEN
       surface%qsol = 0.0_RealType
    ENDIF

    !
    ! Calculate some other surface parameters
    !
    TauTot = SQRT(surface%wind_stress_x**2+surface%wind_stress_y**2)
    surface%Ustar_atmos = SQRT(TauTot/rho_atmos)
    surface%wind_speed = surface%Ustar_atmos/sqrt_cd
    surface%WU_surf = -surface%wind_stress_x/rho_ocean
    surface%WV_surf = -surface%wind_stress_y/rho_ocean

    !GustFactor = 0.5
    !GustFactor = 1.0
    !GustFactor = 1.5
    !GustFactor = sqrt(surface%wind_speed**2 + 0.5**2)
    GustFactor = sqrt(surface%wind_speed**2 + 1.0**2)
    !Gust = surface%wind_speed/GustFactor
    Gust = GustFactor**2 / surface%wind_speed**2
    Gust = 1.0_RealType
    surface%WU_surf = surface%WU_surf*Gust
    surface%WV_surf = surface%WV_surf*Gust

    !
    ! Get Ustar = Ustar_atmos*sqrt(rho_air/rho_sea)
    !
    surface%Ustar = surface%Ustar_atmos*sqrt_rho_air_sea

    surface%Ustar = surface%Ustar*Gust

    !
    ! Get net longwave and solar flux at surface
    !
    total_heat = surface%heat_flux_longwave + &
         surface%heat_flux_latent + surface%heat_flux_sensible
    surface%net_longwave = total_heat * 0.246e-6_RealType
    surface%net_solar = surface%qsol * 0.246e-6_RealType

    !
    ! If required modify the fluxes taking into account diurnal warming
    ! if any
    !
    IF( params%modify_fluxes )THEN
       !
       ! Modify fluxes on the basis of diurnal warming
       !
       CALL Modify_Fluxes( surface, profile )
    END IF

    IF( params%use_langmuir )THEN
       CALL Get_Stokes( surface, profile )
    ENDIF

    !
    ! If first run set tskin to model sst at this point 
    ! With be modified as the code runs
    !
    IF( 1 .eq. istep )THEN
       surface%tskin = surface%model_sst
    ENDIf

    !
    ! Set rain parameters - defaulted to zero
    !
    IF( params%use_evaporation )THEN
       surface%evaporation = surface%heat_flux_latent / &
            ((2.5008E-6_RealType - 2.3E3_RealType * surface%tskin)*rho_ocean)
    ELSE
       surface%evaporation = 0.0_RealType
    ENDIF
    surface%precipitation = 0.0_RealType

    !
    ! Modify salinity flux based on rainfall
    !
    surface%salinity_flux = -profile%salinity(1)*(surface%evaporation-&
         surface%precipitation)

    !
    ! Set read in flux files
    !
    surface%fluxes_init = .TRUE.
    surface%surface_there = .TRUE.

  END SUBROUTINE Get_Fluxes

  !=================================================================
  !
  ! NAME:
  !
  !      Modify_Fluxes
  !
  ! FUNCTION:
  !
  !      Modify surface fluxes based on warming
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Modifies the surface fluxes based on changes in the surface 
  !      temperature caused by Diurnal warming
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Modify_Fluxes(surface,profile)
  !
  ! INPUTS:
  !
  !      profile   - input profile
  !
  ! OUTPUTS:
  !
  ! INPUTS/OUTPUTS:
  !
  !      surface   - structure with surface/model parameters
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  SUBROUTINE Modify_Fluxes( surface, profile )

    TYPE(Surface_type), INTENT(INOUT) :: surface
    TYPE(Profile_type), INTENT(IN) :: profile

    ! Local variables
    REAL(RealType) :: diurnal_warming
    REAL(RealType) :: new_model_sst
    REAL(RealType) :: correction
    REAL(RealType) :: new_diff    
    REAL(RealType) :: old_diff    
    REAL(RealType) :: new_specific_humidity
    REAL(RealType) :: model_specific_humidity
    REAL(RealType) :: new_flux 
    REAL(RealType) :: old_flux
    REAL(RealType) :: diff_flux

    !
    ! Get diurnal warming from current profile
    !
    diurnal_warming = profile%Temp(1)-profile%Temp(profile%pos_reference)
    new_model_sst = surface%model_sst + diurnal_warming

    !
    ! Correct sensible heat flux for DW
    !
    old_diff = surface%model_sst-surface%temp_atmos
    IF( ABS(old_diff) .gt. 0.01_RealType )THEN
       new_diff = new_model_sst-surface%temp_atmos
       correction = new_diff/old_diff
       surface%heat_flux_sensible = surface%heat_flux_sensible * correction
       IF( ABS(surface%heat_flux_sensible) .gt. 10._RealType .and. &
            ((new_diff .gt. 0._RealType .and. surface%heat_flux_sensible .lt. &
            0._RealType) .or. (new_diff .lt. 0._RealType .and. &
            surface%heat_flux_sensible .gt. 0._RealType)) )THEN
          surface%heat_flux_sensible = -surface%heat_flux_sensible
       ENDIF
    ENDIF

    !
    ! Correct latent heat flux for DW
    !
    ! Must first compute saturation specific humidities based on surface
    ! temperatures
    !
    CALL humidity(new_model_sst,surface%pressure,new_specific_humidity)
    CALL humidity(surface%model_sst,surface%pressure,model_specific_humidity)

    old_diff = model_specific_humidity-surface%specific_humidity_atmos
    IF( ABS(old_diff) .gt. 0.01_RealType )THEN
       surface%heat_flux_latent = surface%heat_flux_latent * &
            (new_diff / old_diff)
       IF( ABS(surface%heat_flux_latent) .gt. 10._RealType .and. &
            ((new_diff .gt. 0._RealType .and. &
            surface%heat_flux_latent .lt. 0._RealType) .or. &
            (new_diff .lt. 0._RealType .and. &
            surface%heat_flux_latent .gt. 0._RealType)) )THEN
          surface%heat_flux_latent = -surface%heat_flux_latent
       ENDIF
    ENDIF

    !
    ! Longwave heat flux
    !
    ! Note sst's assumed degree C so convert to K for Stefans-Boltzmann law
    !
    new_flux = 5.67E-8_RealType*(new_model_sst+convert_to_K)**4
    old_flux = 5.67E-8_RealType*(surface%model_sst+convert_to_K)**4

    diff_flux = old_flux - surface%heat_flux_longwave/0.97_RealType
    
    surface%heat_flux_longwave = 0.97*(new_flux - diff_flux)

  END SUBROUTINE Modify_Fluxes

  !=================================================================
  !
  ! NAME:
  !
  !      humidity
  !
  ! FUNCTION:
  !
  !      Get the humidity at the surface of the ocean
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Get the specific humidity at the surface of the ocean
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL humidity(sst,pressure,Q)
  !
  ! INPUTS:
  !
  !      sst      - Sea Surface Temperature
  !      pressure - surface pressure
  !
  ! OUTPUTS:
  !
  !      Q        - output specific humidity
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !
  SUBROUTINE humidity(sst,pressure,Q)

    REAL(RealType), INTENT(IN) :: sst
    REAL(RealType), INTENT(IN) :: pressure
    REAL(RealType), INTENT(OUT) :: Q

    !
    ! humidity in mb
    !
    Q = (1.0007_RealType+3.46E-6_RealType*pressure)*6.1121_RealType * &
         exp(17.502_RealType*sst/(240.97_RealType+sst))
    
    !
    ! Reduce for salinity Kraus 1972 p. 46
    !
    Q = Q*0.98_RealType
    
    ! 
    ! Convert from mb to specific humidity g/kg
    !
    Q = 621.97_RealType*(Q/(pressure-0.378_RealType*Q))

  END SUBROUTINE humidity

  !=================================================================
  !
  ! NAME:
  !
  !      Get_Stokes
  !
  ! FUNCTION:
  !
  !      Routine to estimate the stokes drift
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Comment from original code:
  !      
  !c     Routine to estimate the stokes drift given the significant wave
  !c     height and wave period
  !c
  !c     Approach here for surface stokes drift is following Webb and 
  !c     Fox-Kemper assuming the Pierson and Moskowitz spectrum.
  !c     Assumption of exponential decay with depth is from Janssen
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Get_Stokes(surface,profile)
  !
  ! INPUTS:
  !
  !      surface  - surface structure
  !
  ! OUTPUTS:
  !
  ! INPUTS/OUTPUTS:
  !
  !      profile  - depth profile
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/10/2013
  !      
  !

  SUBROUTINE Get_Stokes( surface, profile )

    TYPE(surface_type), INTENT(IN) :: surface
    TYPE(profile_type), INTENT(INOUT) :: profile

    ! Local variables
    INTEGER(Int4) :: K
    REAL(RealType) :: US_top
    REAL(RealType) :: depth_scale
    REAL(RealType) :: SU

    !
    ! Make sure we have a windspeed
    !
    IF( 0._RealType .lt. surface%wind_speed )THEN
       US_top = 4.289_RealType * PI**3 * surface%significant_wave_height**3/&
            (grav*surface%wave_period**3)
       
       depth_scale = grav/(surface%wind_speed**2)
       
       DO K=1,profile%nlevels
          SU = US_Top * exp(2.0_RealType*depth_scale*profile%zz(K) * &
               profile%Bottom)
          profile%Velocity_U_Stokes(K) = SU*sin(surface%wave_dir*d2r)
          profile%Velocity_V_Stokes(K) = SU*cos(surface%wave_dir*d2r)
       END DO
    ENDIF

  END SUBROUTINE Get_Stokes

END MODULE Get_Data
