!=================================================================
!
! NAME: 
!
!    Get_Parameters
!
! FUNCTION:
!
!    Reads in the parameters from a parameter file
!
! TYPE/LANGUAGE: 
!
!    F90 module
!
! DESCRIPTION:  
!
!    Reads in and parse the information in the input parameter file
!
! ROUTINES TO BE USED EXTERNALLY:
!
!    Read_Parameter_File  - Reads in the parameter file in the parameter 
!                           structure
!    Get_Rad_Bands        - Gets the 9 band radiative absorption coefficients
!
! ROUTINES USED INTERNALLY:
!
!    Get_Parameter        - Reads and parses the file
!    UPCASE               - String uppercase routine
!
! HISTORY:
! 
!   Original version by Jonathan P.D. Mittaz on 9/18/2013
!
!=================================================================
MODULE Get_Parameters

  USE Kinds
  USE Types
  USE Constants
  USE ErrorHandler
  USE Output_Values

  IMPLICIT NONE

  PRIVATE
  PUBLIC :: Read_Parameter_File
  PUBLIC :: Get_Rad_Bands

CONTAINS

  !=================================================================
  !
  ! NAME:
  !
  !      Read_Parameter_File
  !
  ! FUNCTION:
  !
  !      Reads in parameter file and calculates extra parameters
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Reads in parameter file and compares to list of expected parameters
  !      in the Parameter_Type structure.  If present, parses out the relevant
  !      information.  If not present, returns an error
  !
  !      Note that if the Parameter_Type and associated structures in Types.f90
  !      are updated with new parameters then this routine has to be modifies
  !      as well.
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !      Get_Parameter
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Read_Parameter_File( filename, params )
  !
  ! INPUTS:
  ! 
  !      filename  - parameter input file
  !
  ! INPUT/OUTPUTS:
  !
  !      params    - structure containing the parameters read in
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !
  SUBROUTINE Read_Parameter_File( filename, params, Model_Dir, DateString )

    CHARACTER(LEN=*), INTENT(IN) :: filename
    TYPE(parameter_type), INTENT(OUT) :: params
    CHARACTER(LEN=*), INTENT(IN) :: Model_Dir
    CHARACTER(LEN=*), INTENT(IN) :: DateString

    ! Local variables
    INTEGER(Int4) :: Unit
    INTEGER(Int4) :: IOS
    INTEGER(Int4) :: I
    INTEGER(Int4) :: POS

    INTEGER(Int4) :: IntVal
    REAL(RealType) :: RealVal
    LOGICAL :: LogicalVal
    CHARACTER(LEN=256) :: CharVal
    INTEGER :: nParams
    CHARACTER(LEN=256) :: string
    CHARACTER(LEN=256) :: Grid_Stem
    CHARACTER(LEN=256) :: Flux_Stem

    Unit = Get_File_Unit()
    OPEN(UNIT=Unit,FILE=filename,STATUS='OLD',IOSTAT=IOS)
    IF( 0 .ne. IOS )THEN
       CALL out_ERROR('Cannot open parameter file','Read_Parameter_File',&
            'get_parameters.f90')
    ENDIF

    !
    ! Loop round parmeter list to find
    !
    ! Note that the CASE() strings have to agree in case with the structure
    ! parameter_names in Types.f90
    !
    ! In the file to be read in it doesn't matter
    !
    nParams = 0
    DO I=1,nParameters
       CALL Get_Parameter(Unit,TRIM(Parameter_names(I)),&
            TRIM(Parameter_VarType(I)),IntVal,RealVal,LogicalVal,CharVal)
       SELECT CASE (Parameter_names(I))
       CASE('days')
          params%days = RealVal
          nParams = nParams + 1
       CASE('print_freq_hours')
          params%print_freq_hours = RealVal
          nParams = nParams + 1
       CASE('nlevels')
          params%nlevels = IntVal
          nParams = nParams + 1
       CASE('dt1')
          params%dt1 = RealVal
          params%dt2 = params%dt1*2.
          nParams = nParams + 1
       CASE('smooth')
          params%smooth = RealVal
          nParams = nParams + 1
       CASE('FCT')
          params%FCT = RealVal
          nParams = nParams + 1
       CASE('bottom_depth')
          params%bottom_depth = RealVal
          nParams = nParams + 1
       CASE('bottom_salinity')
          params%bottom_salinity = RealVal
          nParams = nParams + 1
       CASE('KMB_init')
          params%KMB_init = RealVal
          nParams = nParams + 1
       CASE('KHB_init')
          params%KHB_init = RealVal
          nParams = nParams + 1
       CASE('shear_enhancement_factor')
          params%shear_enhancement_factor = RealVal
          nParams = nParams + 1
       CASE('wave_enhancement')
          params%wave_enhancement = RealVal
          nParams = nParams + 1
       CASE('surface_abs')
          params%surface_abs = LogicalVal
          nParams = nParams + 1
       CASE('modify_fluxes')
          params%modify_fluxes = LogicalVal
          nParams = nParams + 1
       CASE('use_langmuir')
          !
          ! Note that sometimes the input flux data doesn't have wave
          ! information so we have to turn this off and on on the fly
          ! so we need two variables - one to store the user request, one
          ! to store how the code uses the information
          !
          params%use_langmuir_param = LogicalVal
          params%use_langmuir = LogicalVal
          nParams = nParams + 1
       CASE('exponential_damping')
          params%exponential_damping = LogicalVal
          nParams = nParams + 1
       CASE('use_ascii_input')
          params%use_ascii_input = LogicalVal
          nParams = nParams + 1
       CASE('shear_flag')
          params%shear_flag = LogicalVal
          nParams = nParams + 1
       CASE('wave_breaking')
          params%wave_breaking = LogicalVal
          nParams = nParams + 1
       CASE('use_evaporation')
          params%use_evaporation = LogicalVal
          nParams = nParams + 1
       CASE('File_Turb_Cnts')
          params%File_Turb_Cnts = TRIM(CharVal)
          nParams = nParams + 1
       CASE('Output_File')
          params%Output_File = TRIM(CharVal)
          nParams = nParams + 1
       CASE('test_output')
          params%test_output = LogicalVal
          nParams = nParams + 1
       CASE('Output_File_NetCDF_Format')
          params%Output_File_NetCDF_Format = LogicalVal
          nParams = nParams + 1
       CASE('Use_Solid_Bottom')
          params%Use_Solid_Bottom = LogicalVal
          nParams = nParams + 1
       CASE('Solid_Bottom_Temp')
          params%solid_bottom_temp = RealVal
          nParams = nParams + 1
       CASE('use_skin_temperature')
          params%use_skin_temperature = LogicalVal
          nParams = nParams + 1
       CASE('output_ascii_profiles')
          params%output_ascii_profiles = LogicalVal
          nParams = nParams + 1
       CASE('output_file_profile')
          params%output_file_profile = CharVal
          nParams = nParams + 1
       CASE('langmuir_top_boundary_factor')
          params%langmuir_top_boundary_factor = RealVal
          nParams = nParams + 1
       CASE('reference_depth')
          params%reference_depth = RealVal
          !
          ! Force reference depth to be -ve (below the surface)
          !
          IF( 0 .lt. params%reference_depth )THEN
             params%reference_depth = -params%reference_depth
          ENDIF
          nParams = nParams + 1
       CASE DEFAULT
          WRITE(string,'(''Unknown parameter '',a,'' in parameter file'')')&
               TRIM(Parameter_Names(I))
          CALL out_ERROR(string,'Read_Parameter_File','get_parameters.f90')
       END SELECT
    END DO

    IF( nParams .ne. nParameters )THEN
       CALL out_ERROR('Missing parameters in parameter file',&
            'Read_Parameter_File','get_parameters.f90')
    ENDIF
    
    !
    ! Now look for cases where a particular mode has been set.
    ! Only two currently available - ASCII input file or NetCDF
    ! versions
    !
    IF( params%use_ascii_input )THEN
       CALL Get_Parameter(Unit,'Grid_File_Stem','CHARACTER',&
            IntVal,RealVal,LogicalVal,Grid_Stem)
       CALL Get_Parameter(Unit,'Flux_File_Stem','CHARACTER',&
            IntVal,RealVal,LogicalVal,Flux_Stem)
       params%File_Grid = TRIM(Model_Dir)//'/'//TRIM(DateString)//&
            TRIM(Grid_Stem)
       params%File_Fluxes = TRIM(Model_Dir)//'/'//TRIM(DateString)//&
            TRIM(Flux_Stem)
       CALL Get_Parameter(Unit,'ngrid','INT',&
            IntVal,RealVal,LogicalVal,CharVal)
          params%ngrid = IntVal
       CALL Get_Parameter(Unit,'File_Grid_NX','INT',&
            IntVal,RealVal,LogicalVal,CharVal)
       params%grid_nx = IntVal
       CALL Get_Parameter(Unit,'File_Grid_NY','INT',&
            IntVal,RealVal,LogicalVal,CharVal)
       params%grid_ny = IntVal
       CALL Get_Parameter(Unit,'File_Grid_Min_Lon','REAL',&
            IntVal,RealVal,LogicalVal,CharVal)
       params%grid_min_lon = RealVal
       CALL Get_Parameter(Unit,'File_Grid_Max_Lon','REAL',&
            IntVal,RealVal,LogicalVal,CharVal)
       params%grid_max_lon = RealVal
       CALL Get_Parameter(Unit,'File_Grid_Min_Lat','REAL',&
            IntVal,RealVal,LogicalVal,CharVal)
       params%grid_min_lat = RealVal
       CALL Get_Parameter(Unit,'File_Grid_Max_Lat','REAL',&
            IntVal,RealVal,LogicalVal,CharVal)
       params%grid_max_lat = RealVal
    ENDIF
    IF( params%Output_File_NetCDF_Format )THEN
       params%File_Netcdf = TRIM(Model_Dir)//'/'//TRIM(DateString)
       POS = INDEX(DateString,'.nc')
       IF( 0 .eq. POS )THEN
          params%output_file = TRIM(Model_Dir)//'/'//TRIM(params%output_file)&
               //'_'//TRIM(DateString)//'.nc'
       ELSE
          params%output_file = TRIM(Model_Dir)//'/'//TRIM(params%output_file)&
               //'_'//TRIM(DateString)
       ENDIF
    ELSE
       params%output_file = TRIM(Model_Dir)//'/'//TRIM(params%output_file)&
            //'_'//TRIM(DateString)//'.dat'
    ENDIF

    params%date_there = .FALSE.

    CLOSE(Unit)

    !
    ! Get output frequency
    !
    params%print_freq_loop = INT((3600./params%print_freq_hours)/params%dt1)

  END SUBROUTINE Read_Parameter_File

  !=================================================================
  !
  ! NAME:
  !
  !      Get_Parameter
  !
  ! FUNCTION:
  !
  !      Parses the input line from the parameter file
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Parses a line read in from the parameter file in the basis of the
  !      type of parameter (stored in the structure ParamVar_Type).  Outputs
  !      value to appropriate variable.  Note the code searches for the 
  !      appropriate parameter in the file so it is not order specific.
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Get_Parameter( Unit,Parameter_name,Parameter_VarType,
  !                          IntVal,RealVal,LogicalVal,CharVal)
  !
  ! INPUTS:
  ! 
  !      Unit              - parameter file logical unit no
  !      Parameter_name    - name of parameter to be searched for
  !      Parameter_VarType - Type (Real, Int etc.) of parameter
  !
  ! OUTPUTS:
  !
  !      IntVal             - if vartype==INT output integer value
  !      RealVal            - if vartype==REAL output real value
  !      LogicalVal         - if vartype==LOGICAL output logical value
  !      CharVal            - if vartype==CHAR output string
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !
  SUBROUTINE Get_Parameter(Unit,Parameter_name,Parameter_VarType,&
            IntVal,RealVal,LogicalVal,CharVal)

    INTEGER(Int4), INTENT(IN) :: Unit
    CHARACTER(LEN=*), INTENT(IN) :: Parameter_name
    CHARACTER(LEN=*), INTENT(IN) :: Parameter_VarType
    INTEGER(Int4), INTENT(OUT) :: IntVal
    REAL(RealType), INTENT(OUT) :: RealVal
    LOGICAL, INTENT(OUT) :: LogicalVal
    CHARACTER(LEN=256), INTENT(OUT) :: CharVal

    ! Local variables
    INTEGER :: IOS
    INTEGER :: ipos
    CHARACTER(LEN=256) :: line
    CHARACTER(LEN=256) :: newline
    CHARACTER(LEN=80) :: parameter_name_upcase
    LOGICAL :: parameter_there

    !
    ! Make name uppercase
    !
    parameter_name_upcase = UPCASE(Parameter_name)

    !
    ! Rewind file to get to beginning
    !
    REWIND(Unit)

    !
    ! Loop round and find parameter_names
    !
    parameter_there = .FALSE.
    ReadLoop: DO WHILE(.TRUE.)
       READ(Unit,'(a)',IOSTAT=IOS)line
       IF( 0 .ne. IOS )THEN
          EXIT ReadLoop
       ENDIF
       !
       ! avoid comments
       !
       IF( '#' .ne. line(1:1) .and. ' ' .ne. line(1:1) )THEN
          !
          ! Find '='
          !
          ipos = INDEX(line,'=')
          IF( 0 .eq. ipos )THEN
             CALL out_ERROR('Cannot find = in parameter file','Get_Parameter',&
                  'get_parameters.f90')
          ENDIF
          !
          ! Remove leading/trailing spaces if there
          !
          newline = line(1:ipos-1)
          newline = ADJUSTL(newline)
          newline = TRIM(newline)
          !
          ! Make sure it's uppercase
          !
          newline = UPCASE(newline)
          
          IF( newline .eq. parameter_name_upcase )THEN
             parameter_there = .TRUE.
             IF( 'REAL' .eq. parameter_vartype )THEN
                READ(line(ipos+1:LEN_TRIM(line)),*,IOSTAT=IOS)RealVal
                IF( 0 .ne. IOS )THEN
                   WRITE(newline,'(''Error parsing Real Variable : '',a)',&
                        IOSTAT=IOS)TRIM(Parameter_name)
                   IF( 0 .ne. IOS )THEN
                      CALL out_ERROR('Error making error string',&
                           'Get_Parameter','get_parameters.f90')
                   ENDIF
                   CALL out_ERROR(newline,'Get_Parameter','get_parameters.f90')
                ENDIF
             ELSE IF('INT' .eq. parameter_vartype )THEN
                READ(line(ipos+1:LEN_TRIM(line)),*,IOSTAT=IOS)IntVal
                IF( 0 .ne. IOS )THEN
                   WRITE(newline,'(''Error parsing Int Variable : '',a)',&
                        IOSTAT=IOS)TRIM(Parameter_name)
                   IF( 0 .ne. IOS )THEN
                      CALL out_ERROR('Error making error string',&
                           'Get_Parameter','get_parameters.f90')
                   ENDIF
                   CALL out_ERROR(newline,'Get_Parameter','get_parameters.f90')
                ENDIF
             ELSE IF('LOGICAL' .eq. parameter_vartype )THEN
                newline = ADJUSTL(line(ipos+1:LEN_TRIM(line)))
                IF( ('Y' .eq. newline(1:1)) .or. ('y' .eq. newline(1:1)) )THEN
                   LogicalVal = .TRUE.
                ELSE
                   LogicalVal = .FALSE.
                ENDIF
             ELSE IF('CHARACTER' .eq. parameter_vartype )THEN
                CharVal = ADJUSTL(line(ipos+1:LEN_TRIM(line)))
             ELSE
                CALL out_ERROR('Unknown parameter type','Get_Parameter',&
                     'get_paremeters.f90')
             ENDIF
             EXIT ReadLoop
          ENDIF
       ENDIF
       
    END DO ReadLoop
    
    IF( .not.parameter_there )THEN
       WRITE(newline,'(''Cannot find parameter '',a)',IOSTAT=IOS)&
            TRIM(Parameter_name)
       IF( 0 .ne. IOS )THEN
          CALL out_ERROR('Error making error string','Get_Parameter',&
               'get_parameters.f90')
       ENDIF
       CALL out_ERROR(TRIM(newline),'Get_Parameter','get_parameters.f90')
    ENDIF

  END SUBROUTINE Get_Parameter

  !=================================================================
  !
  ! NAME:
  !
  !      UPCASE
  !
  ! FUNCTION:
  !
  !      Converts a string to uppercase
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Converts a string to uppercase
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      newString = UPCASE(oldString)
  !
  ! INPUTS:
  ! 
  !      oldString  - string to be made uppercase
  !
  ! OUTPUTS:
  !
  ! RETURNS:
  !
  !      newString  - uppercase version of oldString
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !
  PURE FUNCTION UPCASE(Str) RESULT(String)

    CHARACTER(LEN=*), INTENT(IN) :: Str
    CHARACTER(LEN=LEN(Str)) :: String

    ! Local variables
    INTEGER :: I
    INTEGER :: IC

    ! upper and lowercase values
    CHARACTER(LEN=26), PARAMETER :: cap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    CHARACTER(LEN=26), PARAMETER :: low = 'abcdefghijklmnopqrstuvwxyz'

    String = Str
    DO I=1,LEN_TRIM(Str)
       IC = INDEX(low,Str(I:I))
       IF( IC .gt. 0 )String(I:I) = cap(IC:IC)
    END DO

  END FUNCTION UPCASE

  !=================================================================
  !
  ! NAME:
  !
  !      Get_Rad_Bands
  !
  ! FUNCTION:
  !
  !      Gets 9 band radiative absorption coefficients
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Gets 9 band radiative absorption coefficients
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Get_Rad_Bands( rad_bands )
  !
  ! INPUTS:
  ! 
  ! OUTPUTS:
  !
  !      rad_bands  - 9 band coefficients
  !
  ! RETURNS:
  !
  !      newString  - uppercase version of oldString
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !
  SUBROUTINE Get_Rad_Bands( rad_bands )

    TYPE(Rad_Bands_Type), INTENT(OUT) :: rad_bands

    ! Local variables
    INTEGER :: I

    rad_bands%nbands = 9
    IF( rad_bands%nbands .ne. max_nbands )THEN
       CALL out_ERROR('rad_bands%nbands != max_nbands','Get_Rad_Bands',&
            'get_parameters.f90')
    ENDIF

    rad_bands%alpha = (/0.2370_RealType, 0.3600_RealType, 0.1790_RealType, &
         0.0870_RealType, 0.0800_RealType, 0.0250_RealType, &
         0.0250_RealType, 0.0070_RealType, 0.0004_RealType /)
    rad_bands%scl = (/34.840_RealType, 2.2660_RealType, 0.0315_RealType, &
         0.0055_RealType, 8.32E-4_RealType, 1.26E-4_RealType, &
         3.13E-4_RealType,7.82E-4_RealType,1.44E-5_RealType /)

    DO I=1,rad_bands%nbands
       rad_bands%scli(I) = 1.0_RealType/rad_bands%scl(I)
    END DO

    rad_bands%there = .TRUE.

  END SUBROUTINE Get_Rad_Bands

END MODULE Get_Parameters
