!=================================================================
!
! NAME: 
!
!    kantha_clayson_dwmodel
!
! FUNCTION:
!
!    Top level routines that run the Kantha Clayson diurnal warming model
!
! TYPE/LANGUAGE: 
!
!    F90 program
!
! DESCRIPTION:  
!
!    Impliments the Kantha Clayson diurnal warming model.  For details of
!    the model see Kantha & Clayson (1994).  This code based on fortran
!    code supplied by Gary Wick (gary.a.wick@noaa.gov).
!
! ROUTINES TO BE USED EXTERNALLY:
!
! ROUTINES USED INTERNALLY:
!
!    Main             - Top level main routine
!    Skin_Effect      - Calculates the skin effect
!    Smooth_Values    - Smooths the foward, center and backward values
!                       of parameters for numerical stability
!    print_run_time   - Prints out the run time
!    get_no_days      - Gets number of days from 01/01/2013.  Used in 
!                       print_run_time
!
! HISTORY:
! 
!   Original version by Jonathan P.D. Mittaz on 9/18/2013
!
!=================================================================
PROGRAM Kantha_Clayson_dwmodel

  USE Kinds
  USE Types
  USE Constants
  USE Get_Parameters
  USE Get_Data
  USE Profile
  USE Depth_Routines, Get_Depth=>Get_Depth_Scale_Oper
  USE ErrorHandler
  !USE Output_Values
  USE Output_Values,OutputValues=>Write_Output_File_Orig

  IMPLICIT NONE

  !
  ! Call the program
  !
  CALL Main()

CONTAINS

  !=================================================================
  !
  ! NAME:
  !
  !      Main
  !
  ! FUNCTION:
  !
  !      Top level code that runs the Kantha Clayson diurnal warming 
  !      model
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Runs the Kantha Clayson model
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Main( )
  !
  ! INPUT:
  !
  ! INPUT/OUTPUT:
  ! 
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick (NOAA/OER)
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/20/2013
  !      
  !
  SUBROUTINE Main()

    ! Local variables
    INTEGER(Int4) :: Grid_Loop
    INTEGER(Int4) :: i_Time
    INTEGER(Int4) :: I
    INTEGER(Int4) :: end_loop
    INTEGER(Int4) :: TimeValues1(8)
    INTEGER(Int4) :: TimeValues2(8)
    
    REAL(RealType) :: Coriolis

    TYPE(parameter_type) :: params
    TYPE(Turb_Cnts_Type) :: turb_cnts
    TYPE(Rad_Bands_Type) :: rad_bands
    TYPE(Profile_Type) :: profile
    TYPE(Surface_Type) :: surface

    !
    ! Timing information - note this is an instrinsic function to F90
    !
    CALL Date_And_Time( VALUES=TimeValues1 )

    !
    ! Read in static files and open time variable files (fluxes)
    ! Also allocates the profile structure
    !
    CALL Access_Data_Files( params, turb_cnts )

    !
    ! Calculate how many time loops are requested
    !
    end_loop=INT(0.01+params%days*86400./params%dt1)

    !
    ! Now Allocate profile array
    !
    CALL Allocate_Profile(params,profile)

    !
    ! Get rad bands (9 band Morel)
    !
    CALL Get_Rad_Bands( rad_bands )
    
    !
    ! Setup profile
    !
    CALL Get_Depth( params, rad_bands, profile )

    !
    ! Loop over number of grid points
    !
    GridLoop: DO Grid_Loop=1,params%ngrid

       IF( MOD(Grid_Loop,1000) .eq. 0 )THEN
          WRITE(*,*)'Grid ',Grid_Loop,' of ',params%ngrid
       ENDIF

       !
       ! Get surface grid values (init SST, lat/lon )
       !
       !IF( .not. Read_in_Grid( params, surface, Grid_Loop ) )THEN
          !
          ! Looped round all active grid points in the file if
          ! netcdf version read in
          ! If OLD format never returns .FALSE.
          !
       !   EXIT GridLoop
       !ENDIF
       CALL Read_in_Grid( params, surface )

       !
       ! Initialize profile
       !
       CALL Initialize_Profile( params, surface, profile )

       !
       ! Get density
       !
       CALL Profile_Density( profile )

       !
       ! Write header value to binary file in binary file selected
       !
       !CALL Write_To_Output_File( 0, params, surface, profile, .TRUE. )
       CALL OutputValues( 0, params, surface, profile, .TRUE. )

       !
       ! Loop over all required times
       !
       TimeLoop: DO i_Time=1,end_loop

          !
          ! Get fluxes from model data and solar routines
          !
          ! Note that modflux, newflux and getstokes in old code are all
          ! done under Get_Fluxes in new code
          !
          CALL Get_Fluxes(i_Time, params, profile, surface)

          !
          ! Get Temperature and Salinity
          !
          CALL Profile_Get_Temp_Salinity(params, surface, profile, rad_bands)

          !
          ! For stability smooth backward value and move forward value to
          ! center value
          !
          CALL Smooth_Values( profile%nlevels, profile%Temp, &
               profile%Temp_f, profile%Temp_b, params )
          CALL Smooth_Values( profile%nlevels, profile%Salinity, &
               profile%Salinity_f, profile%Salinity_b, params )

          !
          ! Recalculate density
          !
          CALL Profile_Density( profile )

          !
          ! Estimate forward velocity values - taking into account Stokes drift
          ! if requested
          !
          Coriolis = 1.46E-4*SIN(surface%latitude*d2r)

          !
          ! Note removed references to PX, UG and VG that were in 
          ! original code (but set to zero...).
          !
          IF( params%use_langmuir )THEN
             DO I=1,profile%nlevels_minus_1
                profile%Velocity_U_f(I) = profile%Velocity_U_b(I) + &
                     params%dt2*(Coriolis*(profile%Velocity_V(I) + &
                     profile%Velocity_V_Stokes(I)))
                profile%Velocity_V_f(I) = profile%Velocity_V_b(I) - &
                     params%dt2*Coriolis*(profile%Velocity_U(I) + &
                     profile%Velocity_U_Stokes(I))
             END DO
          ELSE
             DO I=1,profile%nlevels_minus_1
                profile%Velocity_U_f(I) = profile%Velocity_U_b(I) + &
                     params%dt2*(Coriolis*(profile%Velocity_V(I)))
                profile%Velocity_V_f(I) = profile%Velocity_V_b(I) - &
                     params%dt2*Coriolis*(profile%Velocity_U(I))
             END DO
          ENDIF
          
          ! 
          ! Get velocity profiles
          !
          CALL Profile_Get_Velocity( params, surface, profile )
          CALL Smooth_Values( profile%nlevels, profile%Velocity_U, &
               profile%Velocity_U_f, profile%Velocity_U_b, params )
          CALL Smooth_Values( profile%nlevels, profile%Velocity_V, &
               profile%Velocity_V_f, profile%Velocity_V_b, params )

          !
          ! Calculate Q (turbluent kinetic energy)
          !
          ! Note this ignores some code in the original which was 
          ! used for testing.  The two routines
          !
          !    PROFQg
          !    PROFQb
          !
          ! are not included here
          !
          CALL Profile_Get_Q( params, surface, turb_cnts, profile )
          CALL Smooth_Values( profile%nlevels, profile%q2, profile%q2_f, &
               profile%q2_b, params )
          CALL Smooth_Values( profile%nlevels, profile%q2l, profile%q2l_f, &
               profile%q2l_b, params )

          !
          ! Calculate skin effect
          !
          CALL Skin_Effect( surface, profile )

          !
          ! Output values
          !
          !CALL Write_To_Output_File( i_Time, params, surface, profile, .FALSE. )
          CALL OutputValues( i_Time, params, surface, profile, .FALSE. )

       END DO TimeLoop

    END DO GridLoop

    !
    ! Deallocate arrays
    !
    CALL Deallocate_Profile(profile)
    CALL Close_Data_Files(params)
    CALL Close_Output_File(params)
    !CALL Free_NetCDF_Store(params)

    !
    ! Timing information at end of run
    !
    CALL Date_And_Time( VALUES=TimeValues2 )

    !
    ! Print out run time
    !
    CALL Print_Run_Time( TimeValues1, TimeValues2 )

  END SUBROUTINE Main

  !=================================================================
  !
  ! NAME:
  !
  !      Skin_Effect
  !
  ! FUNCTION:
  !
  !      Calculates the skin effect to get the skin temperature
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Calculates the skin effect to get the skin temperature
  !
  !     Comments from original code
  !
  !c    This subroutine uses a parameterization of the Donlon type to
  !c    determine the temperature difference at the skin of the ocean.
  !c    The skin temperature is then defined from the difference and the
  !c    shallowest temperature in the model.  Allowances are made
  !c    for different roughness regimes.
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Skin_Effect( profile )
  !
  ! INPUT:
  !
  !      surface - surface structure
  !  
  ! INPUT/OUTPUT:
  !
  !      profile - profile structure
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/6/2013
  !      
  !
  SUBROUTINE Skin_Effect( surface, profile )

    TYPE(Surface_Type), INTENT(INOUT) :: surface
    TYPE(Profile_Type), INTENT(IN) :: profile

    ! Local variables
    REAL(RealType) :: windspeed
    REAL(RealType) :: delta_T

    windspeed = SQRT(surface%Ustar_atmos**2/1.5E-3_RealType)
    IF( surface%Ustar .le. 0.0_RealType )THEN
       surface%tskin = profile%Temp(1)
       RETURN
    ELSE
       delta_T = 0.151129_RealType + 0.468522_RealType * &
            exp(-0.382032_RealType*windspeed)
    ENDIF

    IF( delta_t .gt. -999_RealType )THEN 
       surface%tskin = profile%Temp(1) - delta_T
    ELSE
       surface%tskin = profile%Temp(1)
    ENDIF

  END SUBROUTINE Skin_Effect

  !=================================================================
  !
  ! NAME:
  !
  !      Smooth_Values
  !
  ! FUNCTION:
  !
  !      Smooths backward/current values for stability
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Smooths backward/current values for stability
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Smooth_Values(ndata, Center_Val, Forward_val,
  !                         Back_val, params )
  !
  ! INPUT:
  !
  !      ndata       - number of levels in profile
  !      params      - Input parameters
  !
  ! INPUT/OUTPUT:
  !
  !      Center_Val  - center value (e.g. Temp(I))
  !      Forward_Val - forward value (e.g. Temp_f(I))
  !      Back_Val    - forward value (e.g. Temp_b(I))
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 9/1/2013
  !      
  !
  SUBROUTINE Smooth_Values( ndata, Center_Val, Forward_val, &
       Back_val, params )

    INTEGER(Int4), INTENT(IN) :: ndata
    REAL(RealType), INTENT(INOUT) :: Center_Val(ndata)
    REAL(RealType), INTENT(INOUT) :: Forward_Val(ndata)
    REAL(RealType), INTENT(INOUT) :: Back_Val(ndata)
    TYPE(Parameter_Type), INTENT(IN) :: params

    ! Local variables
    INTEGER :: I

    DO I=1,ndata
       Back_Val(I) = Center_Val(I) + 0.5_RealType*params%smooth* & 
            (Forward_Val(I)+Back_Val(I)-2._RealType*Center_Val(I))
       Center_Val(I) = Forward_Val(I)
    END DO
    
  END SUBROUTINE Smooth_Values

  !=================================================================
  !
  ! NAME:
  !
  !      Print_Run_Time
  !
  ! FUNCTION:
  !
  !      Prints out the total run time
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Calculates and prints the total run time in seconds to compare
  !      with other versions of the Kantha Clayson code
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Print_Run_Time( Time1, Time2 )
  !
  ! INPUT:
  !
  !      Time1    - Start time from Date_And_Time routine
  !      Time2    - End time from Date_And_Time routine
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 10/21/2013
  !      
  !
  SUBROUTINE Print_Run_Time( Time1, Time2 )
    
    INTEGER(Int4), INTENT(IN) :: Time1(8)
    INTEGER(Int4), INTENT(IN) :: Time2(8)

    ! Local variables
    REAL(Dble) :: Days1
    REAL(Dble) :: Days2
    REAL(Dble) :: Time_Diff

    !
    ! Get number of days from 01/01/2013
    !
    CALL Get_No_Days( Time1, Days1 )
    CALL Get_No_Days( Time2, Days2 )

    !
    ! Get time difference - in seconds
    !
    Time_Diff = (Days2-Days1)*86400.0_Dble

    WRITE(*,*)'     ******** Elapsed Time = ',REAL(Time_Diff,KIND=RealType),&
         ' seconds ********'

  END SUBROUTINE Print_Run_Time
    
  !=================================================================
  !
  ! NAME:
  !
  !      Get_No_Days
  !
  ! FUNCTION:
  !
  !      Given a time, gets number of days (including fractional)
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Given a time from Date_and_Time array, calculates the floating 
  !      point number of days from Jan 1st 2013
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Get_No_Days( Time, Days )
  !
  ! INPUT:
  !
  !      Time     - Time array (INTEGER) from Date_And_Time routine
  !  
  ! OUTPUT:
  !
  !      Days     - Number of days (including fractional)
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 10/21/2013
  !      
  !
  SUBROUTINE Get_No_Days( Time, Days )

    INTEGER(Int4), INTENT(IN) :: Time(8)
    REAL(Dble), INTENT(OUT) :: Days

    ! Local variables
    INTEGER(Int4) :: Year

    ! Days in month (non leap year)
    INTEGER(Int4), PARAMETER, DIMENSION(12) :: Month_Day_Number = &
         (/ 0,   31,  59,  90,  120, 151, 181, 212, 243, 273, 304, 334 /)
    ! Days in month (leap year)
    INTEGER(Int4), PARAMETER, DIMENSION(12) :: Month_Day_Number_Leap = &
         (/ 0,   31,  60,  91,  121, 152, 182, 213, 244, 274, 305, 335 /)

    !
    ! Get number of days from Jan 1st 2013
    !
    ! Note Time array is as follows
    !
    !     Time(1)  -  Year
    !     Time(2)  -  Month
    !     Time(3)  -  Day
    !     Time(4)  -  Time difference with UTC (mins)
    !     Time(5)  -  Hour
    !     Time(6)  -  Minute
    !     Time(7)  -  Second
    !     Time(8)  -  Milliseconds
    !
    ! Years before 2013
    !
    days = 0.0_Dble
    DO Year=2013,Time(1)-1
       ! Check if leap year
       IF( (MOD(Year, 4) .eq. 0 .AND. MOD(Year, 100) .ne. 0) .OR. &
            (MOD(Year, 400) .eq. 0) ) THEN
          !
          ! Leap year
          !
          days = days + 366.0_Dble
       ELSE
          !
          ! Not a leap year
          !
          days = days + 365.0_Dble
       ENDIF
    END DO

    !
    ! Now add in days in current year
    !
    IF( (MOD(Time(1), 4) .eq. 0 .AND. MOD(Time(1), 100) .ne. 0) .OR. &
         (MOD(Time(1), 400) .eq. 0) ) THEN
       !
       ! We are in a leap year
       !
       days = days + Month_Day_Number_Leap(Time(2))
    ELSE
       days = days + Month_Day_Number(Time(2))
    ENDIF

    !
    ! Add in days in month - correct for current day
    !
    days = days + Time(3)-1.0_Dble

    !
    ! Add in hours/mins/secs/msecs
    !
    days = days + (Time(5)/24.0_Dble + Time(6)/1400.0_Dble + &
         Time(7)/86400.0_Dble + Time(8)/86400000.0_Dble)

  END SUBROUTINE Get_No_Days

END PROGRAM Kantha_Clayson_dwmodel
