!=================================================================
!
! NAME: 
!
!    Profile
!
! FUNCTION:
!
!    Does all the calculations relating to the profile (depth)
!
! TYPE/LANGUAGE: 
!
!    F90 module
!
! DESCRIPTION:  
!
!    Routines to calculate the temperature, salinity, velocities and 
!    turbulent kinetic energy as a function of depth
!
! ROUTINES TO BE USED EXTERNALLY:
!
!    Profile_Get_Temp_Salinity - get temperature and salinity
!    Profile_Density           - get density as function of depth
!    Profile_Get_Velocity      - get U/V velocities
!    Profile_Get_Q             - get turbulent kinetic energy
!    Allocate_Profile          - allocate arrays needed for profiles
!    Deallocate_Profile        - deallocate arrays
!    Initialize_Profile        - initialize arrays
!
! ROUTINES USED INTERNALLY:
!
!    Profile_Get_Temperature   - gets the temperature profile
!    Profile_Get_Salinity      - gets the salinity profile
!
! HISTORY:
! 
!   Original version by Jonathan P.D. Mittaz on 9/18/2013
!
!=================================================================
MODULE Profile
  USE Kinds
  USE Types
  USE ErrorHandler
  USE Constants
  USE get_parameters

  IMPLICIT NONE

  ! Global variables
  !
  ! Terms for finite difference solution
  ! Maintains values accross routines (temperature/salinity) as well as 
  ! between U/V velocities
  !
  TYPE Finite_Diff_Str
     LOGICAL :: Allocated
     REAL(RealType), ALLOCATABLE :: A(:)
     REAL(RealType), ALLOCATABLE :: C(:)
     REAL(RealType), ALLOCATABLE :: VH(:)
     REAL(RealType), ALLOCATABLE :: VHP(:)
     REAL(RealType), ALLOCATABLE :: TermVal(:)
  END TYPE Finite_Diff_Str

  !
  ! Global variables
  !
  TYPE(Finite_diff_str) :: Finite_diff

  PRIVATE
  PUBLIC :: Profile_Get_Temp_Salinity
  PUBLIC :: Profile_Density
  PUBLIC :: Profile_Get_Velocity
  PUBLIC :: Profile_Get_Q
  PUBLIC :: Allocate_Profile
  PUBLIC :: Deallocate_Profile
  PUBLIC :: Initialize_Profile

CONTAINS

  !=================================================================
  !
  ! NAME:
  !
  !      Profile_Get_Temp_Salinity
  !
  ! FUNCTION:
  !
  !      Returns updated temperature/salinity in the profile
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Gets the updated temperature and salinity profiles
  !
  !      Wrapper for Profile_Get_Potential_Temp and Profile_Get_Salinity
  !      to make sure one follows the other in the correct order and make
  !      sure that updates to A_k, C_k and VH_k are done correctly
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      Profile_Get_Temperature
  !      Profile_Get_Salinity
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Profile_Get_Temp_Salinity(params, surface, profile, rad_bands)
  !
  ! INPUTS:
  ! 
  !      params    - input parameters
  !      rad_bands - N band radiation model parameters
  !
  ! INPUT/OUTPUTS:
  !
  !      surface   - surface parameters
  !      profile   - profiles to be filled with temp and salinity values
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !
  SUBROUTINE Profile_Get_Temp_Salinity(params, surface, profile, rad_bands)

    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(Surface_Type), INTENT(INOUT) :: surface
    TYPE(Profile_Type), INTENT(INOUT) :: profile
    TYPE(Rad_Bands_Type), INTENT(IN) :: rad_bands

    CALL Profile_Get_Temperature(params, surface, profile, rad_bands)
    CALL Profile_Get_Salinity(params, surface, profile)

  END SUBROUTINE Profile_Get_Temp_Salinity

  !=================================================================
  !
  ! NAME:
  !
  !      Profile_Get_Temperature
  !
  ! FUNCTION:
  !
  !      Returns updated potential temperatures in the profile as well as
  !      useful constants.
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Solve potential temperature equation using finite differences
  !      with a leapfrog sceme for stability.
  !
  !      Also sets up A,C,VH and VHP profiles where since we use a centered
  !      difference are
  !
  !      A_k = -2 delta_t (KH+KHB)_(k+1)/((zz_k-zz_(k+1))(z_k-z_(k+1)))
  !      
  !      C_k = -2 delta_t (KH+KHB)_k/((zz_(k-1)-zz_k)(z_k-z_(k+1)))
  !
  !      VH_k = A_K/(A_k + C_k (1-VH_(k-1)) - 1)
  !
  !      VHP_k = (C_k VHP_(k-1)-temp^(n+1)_k+2 delta_t Qt)/
  !                     (A_k + C_k (1-VH_(k-1)) - 1)  
  !      
  !      where
  ! 
  !      k is the profile position
  !      delta_t is the time step
  !      KH is the kinetmatic diffusivity for heat for the turbulent flow
  !      KHB is the kinetmatic diffusivity for heat for the average flow
  !      z is the depth
  !      zz the depth offset by 1/2 a distance step
  !      temp is the temperature
  !      Qt is the source term which accounts for heating by radiation
  !
  !      the potential temperature is then given by
  !
  !      theta_k = VH_k temp_(k+1) + VHP_k
  !
  !      so the code first calculates the VH and VHP terms and then updates
  !      the potential temperatures
  !
  !      Note that since this a 1-dim code only, advection and horizontal
  !      transport terms can be ignored
  !
  !      Note that the 2*dt*Qt term is in the form
  !
  !           delta_SW_Rad / delta_z == (dI/dZ)/(rho Cp) == Qt 
  !
  !      as delta_SW_Rad (profile%swrad) is already scaled (dI/dS)/(rho Cp)
  !
  !      Note also on variable naming.
  !
  !          _f  - variable is forward version
  !          _b  - variable is backward version
  !
  !      Comments from original fortran code
  ! C >>>
  ! C     -------------------------------------------------------------
  ! C     THIS IS A REWRITTEN VERSION SUITABLE FOR BIOLOGICAL & PHOTO-
  ! C     CHEMICAL MODELS. IT INCLUDES PENETRATIVE SOLAR INSOLATION. 
  ! C     MULTIPLE ABSORPTION BANDS ARE CONSIDERED WITH  NONZERO 
  ! C     EXTINCTION SCALES. LAST ONE CAN BE ABSORBED AT THE SURFACE.
  ! C     IF YOU WANT THE LAST COMPONENT ABSORBED AT THE SURFACE
  ! C     MAKE THE DO LOOP GO OVER NBAND-1 (PUT SURABS=1). IF YOU 
  ! C     WANT NONE ABSORBED AT THE SURFACE, MAKE THE DO LOOP GO OVER
  ! C     NBAND (SURABS=0) WITH APPROPRIATE DATA VALUES FOR ALL BANDS.
  ! C
  ! C     THE ALPH AND SCLI ARE FUNCTIONS OF WATER CLARITY. BE SURE
  ! C     TO CHANGE THEM TO FIT THE WATER TYPE.
  ! C     EITHER WAY MAKE SURE ALL ALPH COEFFICIENTS INCLUDING THAT
  ! C     ABSORBED AT THE SURFACE SUM TO UNITY.COEFFICIENTS ALPH
  ! C     (FRACTION OF INSOLATION IN THE BAND) AND SCLI (INVERSE OF
  ! C     EXTINCTION SCALE LENGTH IN INVERSE METERS) NEED TO BE 
  ! C     APPROPRIATELY PRESCRIBED.
  ! C
  ! C     THIS APPLIES ONLY TO SOLAR SW RADIATION. EVERYTHING ELSE
  ! C     AT THE SURFACE IS WFSURF. WFSOLR CAN BE ZERO OR NEGATIVE
  ! C     BUT NOT POSITIVE. 
  ! C
  ! C                                      LAKSHMI KANTHA (7/93)
  ! c
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Profile_Get_Temperature( params, surface, profile, radbands )
  !
  ! INPUTS:
  ! 
  !      params   - Input parameters
  !      radbands - 9 band radiative absorption coefficients
  !
  ! OUTPUTS:
  !
  !      surface  - surface fluxes
  !      profile  - depth profile variables
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !      
  !
  SUBROUTINE Profile_Get_Temperature( params, surface, profile, rad_bands )

    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(Surface_Type), INTENT(INOUT) :: surface
    TYPE(Profile_Type), INTENT(INOUT) :: profile
    TYPE(Rad_Bands_Type), INTENT(IN) :: rad_bands

    ! Local variables
    INTEGER(Int4) :: I
    INTEGER(Int4) :: KI
    REAL(RealType) :: term

    !
    ! Make sure Finite_Diff structure has been allocated
    !
    IF( .not. Finite_diff%allocated )THEN
       CALL out_ERROR('A,C,VH,VHP arrays not allocated',&
            'Profile_Get_Temperature','profile.f90')
    ENDIF

    !
    ! Make sure depth arrays have been assigned
    !
    IF( .not. profile%depth_there )THEN
       CALL out_ERROR('Get_Depth_Scale must be called before this routine',&
            'Profile_Get_Temperature','profile.f90')
    ENDIF
    !
    ! And suface parameters
    !
    IF( .not. surface%surface_there )THEN
       CALL out_ERROR('Get_Surface_Pars must be called before this routine',&
            'Profile_Get_Temperature','profile.f90')
    ENDIF

    !
    ! If surface absorption to be allowed
    !
    IF( params%surface_abs )THEN
       !
       ! Allow extra heating due to radiation
       !
       surface%flux_total = surface%net_longwave + &
            rad_bands%alpha(rad_bands%nbands) * surface%net_solar
    ELSE
       surface%flux_total = surface%net_longwave
    ENDIF

    !
    ! First calculate the source term (Qt)
    ! Note that the solar insolation has already been scaled by (rho Cp) 
    ! (in getflux routines) so
    ! 
    !    delta_SW_Rad / delta_z == (dI/dZ)/(rho Cp) == Qt 
    !
    DO I=1,profile%nlevels
       !
       ! Note that the absorption is already calculated in the depth routine
       ! as it is just a function of z(k) so it is constant and independent
       ! of local conditions
       !
       profile%swrad(I) = profile%swabs(I)*surface%net_solar
    END DO
       
    !
    ! DM term 
    !
    DO I=1,profile%nlevels_minus_1
       !
       ! dm_k = temp_k - 2 delta_t * qt
       !
       profile%dm(I) = profile%temp_b(I) - &
            params%dt2*(profile%swrad(I)-profile%swrad(I+1)) / &
            (profile%delta_z(I)*profile%bottom)
    END DO

    !
    ! Store DM for testing and profile analysis
    !
    profile%dm_temp = profile%dm

    !
    ! Get finite difference contstants
    !
    DO I=2,profile%nlevels_minus_1
       term = -params%dt2*(profile%KH(I)+profile%KH_B(I))/&
            (profile%delta_ZZ(I-1)*profile%bottom_sq)
       !
       ! Note here we are calculating A_(k-1) not A_k which is why term
       ! is constant between A_(K-1) and C_k
       !
       Finite_diff%A(I-1) = term/profile%delta_z(I-1)
       !
       ! here we have C_k
       !
       Finite_diff%C(I) = term/profile%delta_z(I)
    END DO

    ! 
    ! Surface values
    !
    Finite_Diff%VH(1) =  Finite_diff%A(1)/(Finite_diff%A(1)-1._RealType)
    Finite_Diff%VHP(1) =  (-params%dt2*surface%flux_total/&
         (-profile%delta_z(1)*profile%bottom)-profile%dm(1))/&
         (Finite_diff%A(1)-1._RealType)

    !
    ! Rest of profile
    !
    DO I=2,profile%nlevels_minus_2
       term = 1.0_RealType/(Finite_diff%A(I) + Finite_diff%C(I) * &
            (1.0_RealType-Finite_Diff%VH(I-1)) - 1.0_RealType)
       Finite_Diff%VH(I) = Finite_diff%A(I)*term
       Finite_Diff%VHP(I) = (Finite_diff%C(I)*Finite_Diff%VHP(I-1) - &
            profile%dm(I))*term
    END DO
    
    ! For boundary condition 
    !
    ! If solid bottom - force else let float
    !

    IF( params%use_solid_bottom )THEN
       profile%temp_f(profile%nlevels_minus_1) = params%solid_bottom_temp
    ELSE
       !
       ! Just keep bottom consistent with previous step but with accounting
       ! for any advective change.  Note this is used in computing the above
       ! values but then is overridden by reset of new bottom below...
       !
       profile%temp_f(profile%nlevels_minus_1) = &
            profile%temp(profile%nlevels_minus_1)
    ENDIF

    !
    ! Now calculate to bottom 
    !
    DO I=2,profile%nlevels_minus_1
       KI=profile%nlevels-I
       profile%temp_f(KI)=Finite_Diff%VH(KI)*profile%temp_f(KI+1)+&
            Finite_Diff%VHP(KI)
    END DO
    
    !
    ! Set bottom values
    !
    profile%temp_f(profile%nlevels_minus_1) = &
         profile%temp_f(profile%nlevels_minus_2)
    profile%temp_f(profile%nlevels) = profile%temp_f(profile%nlevels_minus_2)

    profile%temperature_there = .TRUE.
    
  END SUBROUTINE Profile_Get_Temperature

  !=================================================================
  !
  ! NAME:
  !
  !      Profile_Get_Salinity
  !
  ! FUNCTION:
  !
  !      Returns updated salinity in the profile
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Solve salinity equation using finite differences
  !      with a leapfrog sceme for stability.
  !
  !      Uses the values of A and C calculated in Profile_Get_Potential_Temp
  !      since profile, KH and KHB have not changed at this point
  !
  !      VHP is different (A,C,VH are the same)
  ! 
  !      VHP_k = (C_k VHP_(k-1)-temp^(n+1)_k)/
  !                     (A_k + C_k (1-VH_(k-1)) - 1)  
  !      
  !      where
  ! 
  !      k is the profile position
  !      A_k a constant dependent on profile (calculated earlier)
  !      C_k a constant dependent on profile (calculated earlier)
  !      delta_t is the time step
  !      temp is the temperature
  !
  !      the salinity is then given by
  !
  !      salinity_k = VH_k salinity_(k+1) + VHP_k
  !
  !      so the code first calculates the VH and VHP terms and then updates
  !      the potential temperatures
  !
  !      Note that since this a 1-dim code only, advection and horizontal
  !      transport terms can be ignored
  !
  !      Note on variable naming.
  !
  !          _f  - variable is forward version
  !          _b  - variable is backward version
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Get_Salinity( params, surface, profile )
  !
  ! INPUTS:
  ! 
  !      params   - Input parameters
  !      surface  - surface fluxes
  !
  ! OUTPUTS:
  !
  !      profile  - depth profile variables
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !      
  !
  SUBROUTINE Profile_Get_Salinity( params, surface, profile )

    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(SURFACE_Type), INTENT(IN) :: surface
    TYPE(Profile_Type), INTENT(INOUT) :: profile

    ! Local variables
    INTEGER(Int4) :: I
    INTEGER(Int4) :: KI
    REAL(RealType) :: term

    IF( .not. Finite_diff%allocated )THEN
       CALL out_ERROR('A,C,VH,VHP arrays not allocated',&
            'Profile_Get_Salinity','profile.f90')
    ENDIF
    IF( .not. profile%depth_there )THEN
       CALL out_ERROR('Get_Depth_Scale must be called before this routine',&
            'Profile_Get_Salinity','profile.f90')
    ENDIF
    IF( .not. profile%temperature_there )THEN
       CALL out_ERROR(&
            'Profile_Get_Temperature must be called before this routine',&
            'Profile_Get_Salinity','profile.f90')
    ENDIF
    IF( .not. surface%surface_there )THEN
       CALL out_ERROR('Get_Surface_Pars must be called before this routine',&
            'Profile_Get_Salinity','profile.f90')
    ENDIF

    !
    ! Salinity is very much like potential temperature except for the 
    ! DM term which is 
    !
    !    dm_k = salinity_k
    !
    ! rather than
    !
    !    dm_k = temp_k - 2 delta_t * qt
    !
    ! DM term 
    !
    DO I=1,profile%nlevels
       profile%dm(I) = profile%salinity_b(I)
    END DO

    !
    ! Note A,C,VH are already calculated in Profile_Get_Potential_Temp
    ! 
    ! Surface values
    !
    Finite_Diff%VHP(1) =  (-params%dt2*surface%salinity_flux/&
         (-profile%delta_z(1)*profile%bottom)-profile%dm(1))/&
         (Finite_diff%A(1)-1.0_RealType)

    !
    ! Rest of profile
    !
    DO I=2,profile%nlevels_minus_2
       term = 1.0_RealType/(Finite_diff%A(I) + Finite_diff%C(I) * &
            (1.0_RealType-Finite_Diff%VH(I-1)) - 1.0_RealType)
       Finite_Diff%VHP(I) = (Finite_diff%C(I)*Finite_Diff%VHP(I-1) - &
            profile%dm(I))*term
    END DO

    !
    ! Bottom salinity is prescribed
    !
    profile%salinity_f(profile%nlevels_minus_1) = params%bottom_salinity

    !
    ! Now calculate to bottom 
    !
    DO I=2,profile%nlevels_minus_1
       KI=profile%nlevels-I
       profile%salinity_f(KI)=Finite_Diff%VH(KI)*profile%salinity_f(KI+1) + &
            Finite_Diff%VHP(KI)
    END DO

    profile%salinity_there = .TRUE.
    
  END SUBROUTINE Profile_Get_Salinity

  !=================================================================
  !
  ! NAME:
  !
  !      Profile_Density
  !
  ! FUNCTION:
  !
  !      Returns updated density in the profile
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Calculates density-1 using salinity and temperature based on
  !      a parameterization
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Profile_Density( profile )
  !
  ! INPUT/OUTPUT:
  ! 
  !      profile  - structure containing depth/profile information
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !      
  !
  SUBROUTINE Profile_Density( profile )
    
    TYPE(Profile_Type), INTENT(INOUT) :: profile

    ! Local variables
    INTEGER(Int4) :: K
    
    IF( .not. profile%initialized )THEN
       CALL out_ERROR(&
            'Profile must be initializedbefore this routine',&
            'Profile_Density','profile.f90')
    ENDIF

    ! 
    ! Get the density via parameterisation of salinity and temperature
    !
    DO K=1,profile%nlevels_minus_1
       profile%rho(K) = 6.76786136E-6_RealType*(profile%salinity(K)**3) - &
            4.8249614E-4_RealType*(profile%salinity(K)**2) + &
            8.14876577E-1_RealType*profile%salinity(K) - 0.22584586E0_RealType

       profile%rho(K) = profile%rho(K) * &
            (1.667E-8_RealType*(profile%temp(K)**3) - &
            8.164E-7_RealType*(profile%temp(K)**2) + &
            1.803E-5_RealType*profile%temp(K))

       profile%rho(K) = profile%rho(K) + 1.0_RealType - &
            1.0843E-6_RealType*(profile%temp(K)**3) + &
            9.8185E-5_RealType*(profile%temp(K)**2) - &
            4.786E-3_RealType*profile%temp(K)

       profile%rho(K) = profile%rho(K) * &
            (6.76786136E-6_RealType*(profile%salinity(K)**3) - &
            4.8249614E-4_RealType*(profile%salinity(K)**2) + &
            8.14876577E-1_RealType*profile%salinity(K) + 3.895414E-2_RealType)

       profile%rho(K) = profile%rho(K) - &
            (profile%temp(K)-3.98_RealType)**2 * &
            (profile%temp(K)+283._RealType)/&
            (503.57_RealType*(profile%temp(K)+67.26_RealType))

       profile%rho(K) = profile%rho(K) * 1.E-3_RealType
    ENDDO
    
  END SUBROUTINE Profile_Density
  
  !=================================================================
  !
  ! NAME:
  !
  !      Profile_Velocity
  !
  ! FUNCTION:
  !
  !      Returns updated the velocity terms (U,V)
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Calculates U and V terms of velocity.  Solves
  !      e.g.
  !     
  !      dU/dt - fV = d/dz((Km+nu)dU/dz)
  !
  !      where
  !
  !      U  - velocity
  !      V  - velocity
  !      f  - 2 Omega sin(phi) - coriolis frequency (Omega - rotation rate of 
  !                              earth, phi - latitude)
  !      Km - mean eddy diffusivity = KM_B in code
  !      nu - fluctuating eddy diffusivity = KM in code
  !      z  - depth scale
  ! 
  !      Using a backward difference it can be shown that
  !
  !      -U^(n+1)_k + U^(n-1)_k + 2 delta_t f V^n + 2 delta_t (
  !
  !        (Km+nu)_k(U^(n+1)_(k-1)-U^(n+1)_k)/(zz_(k-1)-zz_k) - 
  !
  !        (Km+nu)_(k+1)(U^(n+1)_k-U^(n+1)_(k+1))/(zz_k-zz_(k+1))) / 
  !
  !        (z_k - z_(k+1))
  !
  !      and we define
  !
  !      A_k = -2 delta_t (KM+KMB)_(k+1)/((zz_k-zz_(k+1))(z_k-z_(k+1)))
  !      
  !      C_k = -2 delta_t (KM+KMB)_k/((zz_(k-1)-zz_k)(z_k-z_(k+1)))
  !
  !      which gives
  !
  !      -A_k U^(n+1)_(k+1) + (A_k + C_k - 1)U^(n+1)_k - C_k U^(n+1)_(k-1) = 
  !
  !         -U^(n-1)_k - 2 delta_t f V^n
  !      
  !      Same derivation for V component
  !      
  !      Note in the code below the coriolis term has already been added to 
  !      the previous (backward) term previously in the code
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Profile_Velocity( params, surface, profile )
  !
  ! INPUT:
  !
  !      params  - input parameters
  !      surface - surface fluxes
  !
  ! INPUT/OUTPUT:
  ! 
  !      profile - structure containing profile model data
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/14/2013
  !      
  !
  SUBROUTINE Profile_Get_Velocity( params, surface, profile )
    
    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(Surface_Type), INTENT(IN) :: surface
    TYPE(Profile_Type), INTENT(INOUT) :: profile

    ! Local variables
    INTEGER(Int4) :: K
    INTEGER(Int4) :: KI

    REAL(RealType) :: term
    REAL(RealType) :: CBC
    REAL(RealType) :: CBC_Denom
    
    IF( .not. Finite_diff%allocated )THEN
       CALL out_ERROR('A,C,VH,VHP arrays not allocated',&
            'Profile_Velocity','profile.f90')
    ENDIF
    IF( .not. surface%surface_there )THEN
       CALL out_ERROR('Get_Surface_Pars must be called before this routine',&
            'Profile_Density','profile.f90')
    ENDIF
    IF( .not. profile%depth_there )THEN
       CALL out_ERROR('Get_Depth_Scale must be called before this routine',&
            'Profile_Density','profile.f90')
    ENDIF

    ! 
    ! Get A and C parameters
    !
    DO K=2,profile%nlevels_minus_1
       term = -params%dt2 * (profile%KM(K)+profile%KM_B(K))/&
            (profile%delta_zz(K-1)*profile%bottom_sq)
       Finite_Diff%A(K-1) = term/profile%delta_z(K-1)
       Finite_Diff%C(K) = term/profile%delta_z(K)
    END DO

    !
    ! Get VH and VHP parameters
    !
    ! Note that VH is not a function of U or V so can calculated first once
    ! Finite_DifF%TermVal is a store to enable the V component to be done
    ! without recalculation
    !
    ! Momentum flux prescribed at surface
    !
    Finite_Diff%VH(1) = Finite_Diff%A(1)/(Finite_Diff%A(1) - 1.0_RealType)
    Finite_Diff%VHP(1) = (-params%dt2 * surface%WU_surf / &
         (-profile%delta_z(1)*profile%bottom)-profile%Velocity_U_f(1)) / &
         (Finite_diff%A(1)-1.0_RealType)

    !
    ! Loop round levels
    !
    DO K=2,profile%nlevels_minus_2
       Finite_Diff%TermVal(K) = 1.0_RealType/&
            (Finite_Diff%A(K)+Finite_Diff%C(K) * &
            (1._RealType-Finite_Diff%VH(K-1))-1._RealType)
       Finite_Diff%VH(K) = Finite_Diff%A(K)*Finite_Diff%TermVal(K)
       !
       ! VHP for U component - this changes for V component
       !
       Finite_Diff%VHP(K) = (Finite_Diff%C(K)*Finite_Diff%VHP(K-1) - &
            profile%Velocity_U_f(K))*Finite_Diff%TermVal(K)
    END DO

    !
    ! Impliment lower boundary condition
    !
    ! To zero out bottom friction set CBC = 0.
    ! 
    IF( .not. params%use_solid_bottom )THEN
       CBC=0.0_RealType
    ELSE
       !
       ! For case where we have a solid bottom
       !
       ! CBC only needs to be calculated once as it is a function of depth
       ! and U_b and V_b components - same for both U and V.
       !
       CBC=MAX(0.0025_RealType,&
            0.16_RealType/LOG((profile%zz(profile%nlevels_minus_1) - &
            profile%z(profile%nlevels))*profile%bottom/0.01_RealType)**2)
       CBC=CBC*SQRT(profile%Velocity_U_b(profile%nlevels_minus_1)**2 + &
            profile%Velocity_V_b(profile%nlevels_minus_1)**2)
    ENDIF
    !
    ! Denominator of both U and V calculations are the same, so calculate it
    ! once
    ! 
    CBC_Denom = (-CBC*params%dt2/&
         (profile%delta_z(profile%nlevels_minus_1)*profile%bottom)) - 1. - &
         (Finite_Diff%VH(profile%nlevels_minus_2)-1.) * &
         Finite_Diff%C(profile%nlevels_minus_1)

    !
    !  Set value for forward step for U
    !
    profile%Velocity_U_f(profile%nlevels_minus_1) = &
         (Finite_Diff%C(profile%nlevels_minus_1) * &
         Finite_Diff%VHP(profile%nlevels_minus_2) - &
         profile%Velocity_U_f(profile%nlevels_minus_1)) / &
         CBC_Denom

    !
    ! Fill forward values - U component
    !
    DO K=2,profile%nlevels_minus_1
       KI = profile%nlevels - K
       profile%Velocity_U_f(KI)=Finite_Diff%VH(KI)*profile%Velocity_U_f(KI+1)&
            + Finite_Diff%VHP(KI)
    END DO

    !
    ! Now do the same for V component
    !
    ! VH is the same so we don't recalculate it
    !
    Finite_Diff%VHP(1) = (-params%dt2 * surface%WV_surf / &
         (-profile%delta_z(1)*profile%bottom)-profile%Velocity_V_f(1)) / &
         (Finite_diff%A(1)-1.)
    DO K=2,profile%nlevels_minus_2
       !
       ! VHP for U component - this changes for V component
       !
       Finite_Diff%VHP(K) = (Finite_Diff%C(K)*Finite_Diff%VHP(K-1) - &
            profile%Velocity_V_f(K))*Finite_Diff%TermVal(K)
    END DO

    !
    ! Impliment lower boundary condition
    !
    ! To zero out bottom friction set CBC = 0.
    !
    ! Set value for forward step for V
    !
    profile%Velocity_V_f(profile%nlevels_minus_1) = &
         (Finite_Diff%C(profile%nlevels_minus_1) * &
         Finite_Diff%VHP(profile%nlevels_minus_2) - &
         profile%Velocity_V_f(profile%nlevels_minus_1)) / &
         CBC_Denom

    !
    ! Fill forward values - V component
    !
    DO K=2,profile%nlevels_minus_1
       KI = profile%nlevels - K
       profile%Velocity_V_f(KI)=Finite_Diff%VH(KI)*profile%Velocity_V_f(KI+1)&
            + Finite_Diff%VHP(KI)
    END DO

    !
    ! Set value of WV_bottom/WU_bottom
    !
    profile%WU_Bottom = -CBC*profile%Velocity_U_f(profile%nlevels_minus_1)
    profile%WV_Bottom = -CBC*profile%Velocity_V_f(profile%nlevels_minus_1)

  END SUBROUTINE Profile_Get_Velocity

  !=================================================================
  !
  ! NAME:
  !
  !      Profile_Get_Q
  !
  ! FUNCTION:
  !
  !      Solves for q^2 (total kinetic energy) and (q^2 l) where l is the 
  !      turbulent mixing length scale
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Solves the following two equations
  !
  !      dq^2/dt=d/dz(q l Sq d/dz(q^2)) + 2(-<uw>dU/dz-<vw>dV/dz+
  !               beta g <w theta> + beta_s g <w s> - q^3/(B1 l) - nu q^2/z^2)
  !
  !      d(q^2 l)/dt = d/dz(q l Sq d/dz(q^2 l)) + (E1 l)(-<uw>dU/dz-<vw>dV/dz+
  !                    E3(beta g <w theta> + beta_s g <w s>) - 
  !                    E4(q^3/B1)(1+E2(l/(kappa L))^2) - 2 nu q^2 l/z^2
  !
  !      where
  !
  !      q^2    - total kinetic energy
  !      l      - turbulent length scale
  !      L      - distance from a surface (assumes a log-law behavior near sfc)
  !      S,s    - mean salinity, turbulent salinity
  !      U,V    - mean velocities
  !      u,v,w  - turbulent velocities
  !      beta   - -(1/rho)(d rho/d theta)_p
  !      beta_s - -(1/rho)(d rho/d S)_p
  !      kappa  - von Karman constant
  !      nu     - kinematic viscosity
  !      rho    - density
  !
  !      using similar methodology to that described for T,U,V,S
  !
  !      Also derives new values for KM,KH etc.
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Profile_Get_Q( params, surface, turb_cnts, profile )
  ! 
  ! INPUT:
  !
  !      params    - Input parameters
  !      turb_cnts - Turbulent constants (read in and precalculated)
  !
  ! OUTPUT:
  !
  !      surface   - surface fluxes 
  !      profile   - profile
  ! 
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Comments from original code
  !
  !C     ---------------------------------------------------------
  !C       THIS IS A COMPLETELY REVISED VERSION OF M-Y CLOSURE
  !C       IT TAKES INTO ACCOUNT SOME NEW FINDINGS SINCE 1980. 
  !C
  !C     1.  GALPERIN-KANTHA 2 1/4 LEVEL INSTEAD OF M-Y 2 1/2
  !C     2.  PRESSURE-TEMPERATURE COVARIANCES ARE MODELED USING
  !C         FINDINGS OF NCAR LES GROUP (NONZERO C2 & C3)
  !C     3.  BACKGROUND MIXING FUNCTION OF LOCAL FLOW STATE AND 
  !C         NOT A CONSTANT (KMB, KHB NONZERO); STRONG SHEAR-
  !C         INDUCED MIXING IN THE SEASONAL THERMOCLINE.
  !C         PARAMETERIZATION USED BASED ON RICHARDSON NUMBER
  !C         DEEP MIXING FUNCTION OF BUOYANCY FREQUENCY (GARGETT).
  !C         KMB AND KHB ARE SIMPLY ADDED TO KM AND KH.
  !C     4.  EXTINCTION OF TURBULENCE FOR STRONGLY STABLE CASES
  !C         ALLOWED THROUGH A LIMIT ON MINIMUM GH, IMPLEMENTED
  !C         ONLY AWAY FROM THE SURFACE TO PREVENT PREMATURE
  !C         TURBULENCE DAMPING AT THE SURFACE
  !C     5.  IMPLEMENTED IN PROFT: VERY GENERAL SOLAR PENETRATIVE
  !C         HEATING (MULTIBAND) PLUS PROVISIONS TO ABSORB 
  !C         INCIDENT RADIATION AT THE BOTTOM INSTEAD OF IGNORING
  !C         IT (IMPORTANT IN CLEAR SHALLOW WATERS)
  !C     6.  ALL REUSED CONSTANTS PLACED IN PARAMETER STATEMENTS
  !C         TO PREVENT THEM FROM BEING RECALCULATED EVERY STEP
  !C     7.  CODE CONSTRUCTED SO THAT ANY OF THE NEW FEATURES CAN
  !C         BE TURNED OFF AT WILL.
  !C     8.  PROVISION MADE FOR USE OF M-Y 2 1/2 FOR COMPARISONS
  !C     9.  WAVE-MIXING NEAR THE SURFACE ALLOWED
  !C         INDB: 1 FOR FLUX CONDITION AT THE TOP, 0 FOR THE USUAL
  !C         FLUXE: ENERGY FLUX AT THE TOP DUE TO BREAKING WAVES
  !C         ZEENOT: TURBULENCE SCALE AT THE TOP DUE TO BREAKING WAVES
  !C    10.  ENERGY INPUT FROM LANGMUIR CIRCULATION ALLOWED
  !C         INDL: 1 FOR LAGMUIR INPUT, 0 FOR THE USUAL
  !C         US(K), VS(K) DENOTE STOKES DRIFT AT VARIOUS DEPTHS
  !C    11.  THE CODE IS DESIGNED FOR TESTING VARIOUS ALTERNATIVES
  !C         IN TURBULENCE MODELING. CAUTION IS NEEDED IN ITS USE
  !C         NO THOUGHT HAS BEEN GIVEN TO EFFICIENCY OF EXECUTION
  !C         A STREAMLINED VERSION IS USED IN THE 3-D MODEL
  !C
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/20/2013
  !      
  SUBROUTINE Profile_Get_Q( params, surface, turb_cnts, profile )

    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(Surface_Type), INTENT(INOUT) :: surface
    TYPE(Turb_Cnts_Type), INTENT(IN) :: turb_cnts
    TYPE(Profile_Type), INTENT(INOUT) :: profile

    ! Local variables
    INTEGER :: K
    INTEGER :: KI
    REAL(RealType) :: Term
    REAL(RealType) :: Term2
    REAL(RealType) :: Term3
    REAL(RealType) :: du_dz
    REAL(RealType) :: dv_dz
    REAL(RealType) :: dus_dz
    REAL(RealType) :: dvs_dz
    REAL(RealType) :: q2lmax
    REAL(RealType) :: z_in
    REAL(RealType) :: Velocity_Langmuir
    REAL(RealType) :: ensqr

    !
    ! Code is for 2 1/4 case NOT 2 1/2 case
    !
    ! Setup A and C coefficients
    !
    DO K=2,profile%nlevels_minus_1
       Term = profile%delta_ZZ(K-1)*profile%bottom_sq
       Finite_Diff%A(K) = -0.5*params%dt2*(profile%KQ(K+1)+profile%KQ(K)+&
            profile%KH_B(K+1)+profile%KH_B(K))/(Term*profile%delta_z(K))
       Finite_Diff%C(K) = -0.5*params%dt2*(profile%KQ(K-1)+profile%KQ(K)+&
            profile%KH_B(K-1)+profile%KH_B(K))/(Term*profile%delta_z(K-1))
    END DO

    !
    ! The following section solves
    !
    !    2 delta_t d(K_Q dq2/dz)/dz - q2 - 2 delta_t 2 q^3/(l B1) = q2_b
    !
    ! Surface boundary condition
    !
    !
    IF( params%Use_Langmuir )THEN
       Velocity_Langmuir = (SQRT(profile%Velocity_U_Stokes(1)**2 + &
            profile%Velocity_V_Stokes(1)**2)*(surface%Ustar**2))**(1./3.)
    ELSE
       Velocity_Langmuir = 0.0_RealType
    ENDIF

    !
    ! Upper boundary condition
    !
    Finite_Diff%VH(1) = 0.0_RealType
    Finite_Diff%VHP(1)=turb_cnts%total_KE_const*(surface%Ustar**2)
    IF( params%wave_breaking )THEN
       Finite_Diff%VHP(1) = Finite_Diff%VHP(1)*&
         (1.0_RealType+3.0_RealType*EMM*BEE*params%wave_enhancement)**(2./3.)
    ENDIF
    !
    ! Note this is different form original code in that the KE constant
    ! has beend added to get velocity squared to an energy
    !
    IF( params%Use_Langmuir )THEN
        Finite_Diff%VHP(1) = Finite_Diff%VHP(1) + &
             params%langmuir_top_boundary_factor*(Velocity_Langmuir**2)
    ENDIF

    !
    ! Bottom boundary condition
    !
    ! Note this has changed from original code where there was a 
    ! SQRT infront of the velocity squared terms
    !
    profile%q2_f(profile%nlevels) = (profile%WU_Bottom**2 + &
         profile%WV_Bottom**2)*turb_cnts%total_KE_Const

    ! 
    ! Setup some of the required values
    !
    DO K=2,profile%nlevels_minus_1
       z_in = 1./(profile%delta_ZZ(K-1)*profile%bottom)
       !
       ! g drho/dz = rho x N**2 where N is the local buoyancy frequency
       !                          or Brunt-Vaisala frequency
       !
       profile%g_drho_dz(K) = grav * (profile%rho(K-1)-profile%rho(K))*z_in
       !
       ! du/dz and dv/dz terms
       !
       du_dz=(profile%Velocity_U(K-1)-profile%Velocity_U(K))*z_in
       dv_dz=(profile%Velocity_V(K-1)-profile%Velocity_V(K))*z_in
       !
       ! Local gradient Richardson number (Large et al. 1994)
       !
       profile%Richardson_no(K) = -profile%g_drho_dz(K)/&
            ((du_dz**2+dv_dz**2+SMALL)*params%shear_enhancement_factor)
       profile%Richardson_no(K) = MIN(profile%Richardson_no(K),0.7)
       !
       ! Production variables
       !
       ! SPROD in old code
       !
       profile%Shear_prod(K) = profile%KM(K)*(du_dz**2 + dv_dz**2) * &
            params%shear_enhancement_factor
       !
       ! BPROD in old code
       !
       profile%Buoyancy_prod(K) = profile%KH(K)*profile%g_drho_dz(K)

       !
       ! CPROD in old code
       !
       ! If Langmiur circulation included
       !
       IF( params%Use_Langmuir )THEN
          !
          ! Stokes drift components
          !
          dus_dz=(profile%Velocity_U_Stokes(K-1) - &
               profile%Velocity_U_Stokes(K))*z_in
          dvs_dz=(profile%Velocity_V_Stokes(K-1) - &
               profile%Velocity_V_Stokes(K))*z_in
          profile%Shear_prod_stokes(K) = profile%KM(K)*&
               (du_dz*dus_dz+dv_dz*dvs_dz)
       ELSE
          profile%Shear_prod_stokes(K) = 0.0_RealType
       ENDIF
       profile%q2_b(K) = ABS(profile%q2_b(K))
       profile%q2l_b(K) = ABS(profile%q2l_b(K))
       !
       ! Eddy dissipation term - was DTEF in old code
       !
       profile%Eddy_diss_q2l(K) = profile%q2_b(K) * &
            SQRT(profile%q2_b(K)) / (turb_cnts%B1*profile%q2l_b(K)+EPS)
       !
       ! Get total prod
       ! 
       profile%prod(K) = profile%Shear_prod(K) + profile%Buoyancy_prod(K) + &
            profile%Shear_prod_stokes(K)


       !
       ! Get VH, VHP
       !
       Finite_Diff%VHP(K) = 1.0_RealType/(Finite_Diff%A(K)+&
            Finite_Diff%C(K)*(1.0-Finite_Diff%VH(K-1))-&
            (2.0_RealType*params%dt2*profile%Eddy_Diss_q2l(K)+1.0_RealType))
       Finite_Diff%VH(K)=Finite_Diff%A(K)*Finite_Diff%VHP(K)
       Finite_Diff%VHP(K)=(-2.0_RealType*params%dt2*profile%prod(K)+&
            Finite_Diff%C(K)*Finite_Diff%VHP(K-1)-&
            profile%q2_b(K))*Finite_Diff%VHP(K)
    END DO

    !
    ! Now get solution to forward time
    !
    DO K=1,profile%nlevels_minus_1
       KI = profile%nlevels-K
       profile%q2_f(KI) = Finite_Diff%VH(KI)*profile%q2_f(KI+1) + &
            Finite_diff%VHP(KI)
    END DO

    !
    ! Now solve the equation
    !
    !    2 delta_t d(K_Q d(q2 l)/dz)/dz - (q2 l) - 2 delta_t q^3/(B1) = 
    !       (q2 l)_b
    !
    ! Upper boundary conditions
    !
    surface%breakwave_turb_scale=MAX(surface%breakwave_turb_scale,0.01_RealType)
    IF( params%wave_breaking )THEN
       profile%L(1)=BEE*surface%breakwave_turb_scale
    ELSE
       profile%L(1)=Kappa*surface%breakwave_turb_scale
    ENDIF

    Finite_Diff%VH(1) = 0.0_RealType
    Finite_Diff%VHP(1) = profile%q2_f(1) * profile%L(1)

    !
    ! Lower boundary conditions
    !
    profile%q2l_f(profile%nlevels) = 0.0_RealType
    profile%L(profile%nlevels) = 0.0_RealType

    !
    ! Get A,C,VH,VHP
    !
    Term2 = profile%bottom * kappa
    Term3 = (surface%breakwave_turb_scale/profile%bottom)
    DO K=2,profile%nlevels_minus_1
       Term = profile%delta_zz(K-1)*profile%bottom_sq
       Finite_Diff%A(K) = -0.5_RealType * &
            params%dt2*(turb_cnts%E4*turb_cnts%SLSQ* &
            (profile%KQ(K+1)+profile%KQ(K))+profile%KH_b(K+1)+profile%KH_b(K))/&
            (Term*profile%delta_z(K))
       Finite_Diff%C(K) = -0.5_RealType * &
            params%dt2*(turb_cnts%E4*turb_cnts%SLSQ* &
            (profile%KQ(K-1)+profile%KQ(K))+profile%KH_b(K-1)+profile%KH_b(K))/&
            (Term*profile%delta_z(K-1))       
       !
       ! get DTEF (now callled Eddy_Diss_q2l) and VH/VHP
       !
       profile%Eddy_Diss_q2l(K) = profile%Eddy_Diss_q2l(K)*(1.+turb_cnts%E2*&
            ((1.0_RealType/ABS(-profile%z(K)+profile%z(1)+Term3))*&
            profile%L(K)/Term2)**2)
       Finite_Diff%VHP(K) = 1.0_RealType/(Finite_Diff%A(K) + Finite_Diff%C(K)*&
            (1.0_RealType-Finite_Diff%VH(K-1)) - &
            (params%dt2* profile%Eddy_Diss_q2l(K)+1.0_RealType))
       Finite_Diff%VH(K) = Finite_Diff%A(K) * Finite_Diff%VHP(K)
       Finite_Diff%VHP(K) = (-params%dt2*profile%L(K)*&
            (turb_cnts%E1*profile%Shear_prod(K) + &
            turb_cnts%E6*profile%Shear_prod_stokes(K) +&
            turb_cnts%E3*profile%Buoyancy_prod(K)) +&
            Finite_Diff%C(K)*Finite_Diff%VHP(K-1) - profile%q2l_b(K)) * &
            Finite_Diff%VHP(K)
    END DO

    !
    ! Get (q2 l) forward time
    !
    DO K=1,profile%nlevels_minus_1
       KI=profile%nlevels-K
       profile%q2l_f(KI)=Finite_Diff%VH(KI)*profile%q2l_f(KI+1) + &
            Finite_Diff%VHP(KI)
    END DO

    !
    ! Apply bounds to values
    !
    DO K=2,profile%nlevels_minus_1
       IF( (profile%q2_f(K) .le. SMALL) .or. (profile%q2l_f(K) .le. SMALL) )THEN
          profile%q2_f(K) = SMALL
          profile%q2l_f(K) = SMALL
       ENDIF
       ensqr = -profile%g_drho_dz(K)
       IF( ensqr .gt. 0. )THEN
          q2lmax = profile%q2_f(K)*0.53_RealType*SQRT(profile%q2_f(K)/ensqr)
          profile%q2l_f(K) = MIN(profile%q2l_f(K),q2lmax)
       ENDIF
    END DO

    !
    ! Solve for KM and KH
    !
    DO K=2,profile%nlevels_minus_1
       profile%L(K) = profile%q2l_f(K)/(profile%q2_f(K)+EPS)
       profile%qli(K) = profile%L(K)**2/profile%q2_f(K)
       Term = 1.0_RealType/(profile%delta_ZZ(K-1)*profile%bottom)
       du_dz = (profile%Velocity_U(K-1)-profile%Velocity_U(K))*Term
       dv_dz = (profile%Velocity_V(K-1)-profile%Velocity_V(K))*Term
       profile%Richardson_no(K) = -profile%g_drho_dz(K)/&
            ((du_dz**2+dv_dz**2+SMALL)*&
            params%shear_enhancement_factor)
       profile%Richardson_no(K) = MIN(profile%Richardson_no(K),0.7_RealType)
       !
       ! Constant background mixing
       !
       profile%KM_b(K) = params%KMB_init
       profile%KH_b(K) = params%KHB_init
       
       IF (ABS(du_dz) .lt. EPS) du_dz = EPS
       IF (ABS(dv_dz) .lt. EPS) dv_dz = EPS

       !
       ! Level 2 1/4
       !
       profile%GH(K) = profile%QLI(K)*profile%g_drho_dz(K)
       profile%GH(K) = MAX(profile%GH(K),turb_cnts%GH_min)
       profile%GH(K) = MIN(profile%GH(K),turb_cnts%GH_max)

    END DO

    !
    ! Setup boundary conditions
    !
    profile%GH(1) = 0.0_RealType
    profile%GH(profile%nlevels) = 0.0_RealType
    profile%Richardson_no(1) = profile%Richardson_no(2)
    profile%Richardson_no(profile%nlevels) = &
         profile%Richardson_no(profile%nlevels_minus_1)
    profile%KH_b(1) = 0.0_RealType
    profile%KM_b(1) = 0.0_RealType
    profile%KH_b(profile%nlevels) = 0.0_RealType
    profile%KM_b(profile%nlevels_minus_1) = 0.0_RealType

    !
    ! Do 2 1/4 case
    !
    DO K=1,profile%nlevels
       profile%SH(K) = turb_cnts%CF1/(1.0_RealType-turb_cnts%CF2*profile%GH(K))
       profile%SM(K) = (turb_cnts%CF3+&
            turb_cnts%CF4*profile%SH(K)*profile%GH(K))/&
            (1.0_RealType-turb_cnts%CF5*profile%GH(K))
    END DO

    DO K=1,profile%nlevels
       IF( params%shear_flag )THEN
          IF( profile%q2(K) .lt. 1.e-6_RealType )THEN
             profile%KM_b(K) = params%KMB_init
             profile%KH_b(K) = params%KHB_init
             IF( profile%Richardson_no(K) .le. 0.0_RealType )THEN
                profile%KM_b(K) = profile%KM_b(K) + 5.E-3_RealType
                profile%KH_b(K) = profile%KH_b(K) + 5.E-3_RealType
             ENDIF
             IF( profile%Richardson_no(K) .gt. 0.0_RealType .and. &
                  profile%Richardson_no(K) .lt. 0.7_RealType )THEN
                profile%KM_b(K) = profile%KM_b(K) + &
                     5.E-3_RealType*(1.0_RealType - &
                     (profile%Richardson_no(K)/0.7_RealType)**2)**3
                profile%KH_b(K) = profile%KH_b(K) + &
                     5.E-3_RealType*(1. - &
                     (profile%Richardson_no(K)/0.7_RealType)**2)**3
             ENDIF
          ENDIF
       ENDIF

       profile%KN(K) = profile%L(K)*SQRT(profile%q2(K))
       IF( params%exponential_damping )THEN
          !
          ! Surface damping
          !
          IF( surface%flux_total .lt. 0.0_RealType )THEN
             profile%KN(K) = profile%KN(K)*(1.-EXP(surface%Ustar * &
                  profile%Z(K)*profile%bottom/(27.0_RealType*Umol)))
             profile%KM_b(K) = profile%KM_b(K)*(1.-EXP(surface%Ustar * &
                  profile%Z(K)*profile%bottom/(27.0_RealType*Umol)))
             profile%KH_b(K) = profile%KH_b(K)*(1.-EXP(surface%Ustar * &
                  profile%Z(K)*profile%bottom/(27.0_RealType*Umol)))
          ELSE
             profile%KN(K) = profile%KN(K)*(1.0_RealType-EXP(surface%Ustar * &
                  profile%Z(K)*profile%bottom/(5.0_RealType*Umol)))
             profile%KM_b(K) = profile%KM_b(K) * &
                  (1.0_RealType-EXP(surface%Ustar * &
                  profile%Z(K)*profile%bottom/(5.0_RealType*Umol)))
             profile%KH_b(K) = profile%KH_b(K) * &
                  (1.0_RealType-EXP(surface%Ustar * &
                  profile%Z(K)*profile%bottom/(5.0_RealType*Umol)))
          ENDIF
       ENDIF

       profile%KQ(K)=profile%KN(K)*SQ*(1-params%FCT)*&
            profile%SH(K) + profile%KQ(K)*params%FCT
       profile%KM(K)=profile%KN(K)*profile%SM(K)*(1.0_RealType-params%FCT) + &
            profile%KM(K)*params%FCT
       profile%KH(K)=profile%KN(K)*profile%SH(K)*(1.0_RealType-params%FCT) + &
            profile%KH(K)*params%FCT

    END DO

  END SUBROUTINE Profile_Get_Q

  !=================================================================
  !
  ! NAME:
  !
  !      Allocate_Profile
  !
  ! FUNCTION:
  !
  !      Allocated the profile structure
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Allocated profile structure
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Allocate_Profile( params, profile )
  !
  ! INPUT:
  !
  !      params   - input parameters
  !
  ! INPUT/OUTPUT:
  ! 
  !      profile  - Profile structure to be allocated
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/20/2013
  !      
  !
  SUBROUTINE Allocate_Profile( params, profile )

    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(Profile_Type), INTENT(INOUT) :: profile

    !
    ! Local variables
    !
    INTEGER(Int4) :: STAT

    !
    ! Set profile parameters
    !
    profile%nlevels = params%nlevels
    profile%nlevels_minus_1 = params%nlevels-1
    profile%nlevels_minus_2 = params%nlevels-2
    profile%bottom = params%bottom_depth
    profile%bottom_sq = params%bottom_depth**2

    !
    ! Allocate Finitw_Diff struct which is a local global variable
    !
    ! First check to see if A,C,VH,VHP variables have been allocated
    ! and deallocate if that is true
    !
    IF( ALLOCATED(Finite_diff%A) )THEN
       DEALLOCATE(Finite_diff%A,Finite_diff%C,Finite_diff%TermVal,&
            Finite_diff%VH,Finite_diff%VHP,STAT=STAT)
       IF( 0 .ne. STAT )THEN
          CALL out_ERROR('Cannot deallocate finite_diff structures',&
               'Profile_Get_Potential_Temp','profile.f90')
       ENDIF
    ENDIF

    ALLOCATE(Finite_diff%A(profile%nlevels),&
         Finite_diff%C(profile%nlevels),&
         Finite_Diff%TermVal(profile%nlevels),&
         Finite_Diff%VH(profile%nlevels),&
         Finite_Diff%VHP(profile%nlevels),&
         STAT=STAT)
    IF( 0 .ne. STAT )THEN
       CALL out_ERROR('Cannot Allocate finite_diff structures',&
            'Profile_Get_Potential_Temp','profile.f90')
    ENDIF

    Finite_diff%allocated = .TRUE.

    !
    ! Make sure that profile structure is deallocated first
    !
    IF( ALLOCATED(profile%z) )THEN
       DEALLOCATE(profile%z,profile%zz,profile%delta_z,&
         profile%delta_zz,profile%swabs,&
         profile%swrad,profile%dm,&
         STAT=STAT)
       IF( 0 .ne. STAT )THEN
          CALL out_ERROR('Cannot deallocate depth arrays','Allocate_Profile',&
               'profile.f90')
       ENDIF
    ENDIF

    !
    ! Setup profile structure and allocate arrays
    !
    ALLOCATE(profile%z(profile%nlevels),&
         profile%zz(profile%nlevels),&
         profile%delta_z(profile%nlevels),&
         profile%delta_zz(profile%nlevels),&
         profile%rho(profile%nlevels),&
         profile%swabs(profile%nlevels),&
         profile%swrad(profile%nlevels),&
         profile%dm_temp(profile%nlevels),&
         profile%dm(profile%nlevels),&
         profile%Temp(profile%nlevels),&
         profile%Temp_f(profile%nlevels),&
         profile%Temp_b(profile%nlevels),&
         profile%Salinity(profile%nlevels),&
         profile%Salinity_f(profile%nlevels),&
         profile%Salinity_b(profile%nlevels),&
         profile%Velocity_V(profile%nlevels),&
         profile%Velocity_V_f(profile%nlevels),&
         profile%Velocity_V_b(profile%nlevels),&
         profile%Velocity_U(profile%nlevels),&
         profile%Velocity_U_f(profile%nlevels),&
         profile%Velocity_U_b(profile%nlevels),&
         profile%Velocity_U_Stokes(profile%nlevels),&
         profile%Velocity_V_Stokes(profile%nlevels),&
         profile%q2(profile%nlevels),&
         profile%q2_f(profile%nlevels),&
         profile%q2_b(profile%nlevels),&
         profile%q2l(profile%nlevels),&
         profile%q2l_f(profile%nlevels),&
         profile%q2l_b(profile%nlevels),&
         profile%qli(profile%nlevels),&
         profile%L(profile%nlevels),&
         profile%g_drho_dz(profile%nlevels),&
         profile%Richardson_no(profile%nlevels),&
         profile%Shear_prod(profile%nlevels),&
         profile%Buoyancy_prod(profile%nlevels),&
         profile%Shear_prod_stokes(profile%nlevels),&
         profile%Eddy_Diss_q2l(profile%nlevels),&
         profile%prod(profile%nlevels),&
         profile%KH(profile%nlevels),&
         profile%KH_B(profile%nlevels),&
         profile%KM(profile%nlevels),&
         profile%KM_B(profile%nlevels),&
         profile%KQ(profile%nlevels),&
         profile%KN(profile%nlevels),&
         profile%GH(profile%nlevels),&
         profile%SH(profile%nlevels),&
         profile%SM(profile%nlevels),&
         STAT=STAT)
    IF( 0 .ne. STAT )THEN
       CALL out_ERROR('Cannot allocate depth/profile arrays',&
            'Allocate_Profile','profile.f90')
    ENDIF

    profile%profile_allocated = .TRUE.

  END SUBROUTINE Allocate_Profile
  
  !=================================================================
  !
  ! NAME:
  !
  !      Deallocate_Profile
  !
  ! FUNCTION:
  !
  !      Deallocated the profile structure
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Deallocated profile structure
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !      out_ERROR
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Dellocate_Profile( profile )
  !
  ! INPUT/OUTPUT:
  ! 
  !      profile  - Profile structure to be deallocated
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/20/2013
  !      
  !
  SUBROUTINE Deallocate_Profile( profile )

    TYPE(Profile_Type), INTENT(INOUT) :: profile

    !
    ! Local variables
    !
    INTEGER(Int4) :: STAT

    !
    ! Deallocate Finite_Diff structure
    !
    IF( ALLOCATED(Finite_diff%A) )THEN
       DEALLOCATE(Finite_diff%A,Finite_diff%C,Finite_diff%TermVal,&
            Finite_diff%VH,Finite_diff%VHP,STAT=STAT)
       IF( 0 .ne. STAT )THEN
          CALL out_ERROR('Cannot deallocate finite_diff structures',&
               'Profile_Get_Potential_Temp','profile.f90')
       ENDIF
       Finite_diff%allocated = .FALSE.
    ENDIF

    !
    ! Make sure that profile structure is allocated 
    !
    IF( ALLOCATED(profile%z) )THEN
       DEALLOCATE(profile%z,profile%zz,profile%delta_z,profile%swabs,&
         profile%swrad,profile%dm,profile%dm_temp,profile%Temp,profile%Temp_f,&
         profile%Temp_b,profile%Salinity,profile%Salinity_f,&
         profile%Salinity_b,profile%Velocity_V,profile%Velocity_V_f,&
         profile%Velocity_V_b,profile%Velocity_U,profile%Velocity_U_f,&
         profile%Velocity_U_b,profile%Velocity_U_Stokes,&
         profile%Velocity_V_Stokes,profile%q2,profile%q2_f,&
         profile%q2_b,profile%q2l,profile%q2l_f,profile%q2l_b,&
         profile%qli,profile%L,profile%g_drho_dz,&
         profile%Richardson_no,profile%Shear_prod,profile%Buoyancy_prod,&
         profile%Shear_prod_stokes,profile%Eddy_Diss_q2l,&
         profile%prod,profile%KH,profile%KH_B,profile%KM,profile%KM_B,&
         profile%KQ,profile%KN,profile%GH,profile%SH,profile%SM,&
         STAT=STAT)
       IF( 0 .ne. STAT )THEN
          CALL out_ERROR('Cannot deallocate depth/profile arrays',&
               'Deallocate_Profile','profile.f90')
       ENDIF
       profile%nlevels = 0
       profile%nlevels_minus_1 = 0
       profile%nlevels_minus_2 = 0
       profile%bottom = 0
       profile%temperature_there = .FALSE.
       profile%salinity_there = .FALSE.
       profile%depth_there = .FALSE.
    ENDIF

    profile%profile_allocated = .FALSE.

  END SUBROUTINE Deallocate_Profile
  
  !=================================================================
  !
  ! NAME:
  !
  !      Initialize_Profile
  !
  ! FUNCTION:
  !
  !      Initializes the profile structure
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Initializes the profile structure
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !      CALL Initialize_Profile( params, surface, profile )
  !
  ! INPUT:
  !
  !      params   - Input parameters 
  !      surface  - Surface structure containing the initial SST
  !
  ! INPUT/OUTPUT:
  ! 
  !      profile  - Profile structure to be deallocated
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Original F90 version by Jonathan P.D. Mittaz 8/20/2013
  !      
  !
  SUBROUTINE Initialize_Profile( params, surface, profile )

    TYPE(Parameter_Type), INTENT(IN) :: params
    TYPE(Surface_Type), INTENT(IN) :: surface
    TYPE(Profile_Type), INTENT(INOUT) :: profile

    ! Local variables
    INTEGER :: I

    DO I=1,profile%nlevels

       profile%Velocity_U_b(I) = 0.0_RealType
       profile%Velocity_U(I) = profile%Velocity_U_b(I)
       profile%Velocity_U_f(I) = profile%Velocity_U(I)

       profile%Velocity_V_b(I) = 0.0_RealType
       profile%Velocity_V(I) = profile%Velocity_V_b(I)
       profile%Velocity_V_f(I) = profile%Velocity_V(I)

       profile%Temp_b(I) = surface%sstinit
       profile%Temp(I) = profile%Temp_b(I)
       profile%Temp_f(I) = profile%Temp(I)

       profile%Salinity_b(I) = params%bottom_salinity
       profile%Salinity(I) = profile%Salinity_b(I)
       profile%Salinity_f(I) = profile%Salinity(I)

       profile%q2_b(I) = 1.0e-5_RealType
       profile%q2(I) = profile%q2_b(I)
       profile%q2_f(I) = profile%q2(I)

       profile%q2l_b(I) = SQRT(profile%q2(I))*Umol*0.1_RealType
       profile%q2l(I) = profile%q2l_b(I)
       profile%q2l_f(I) = profile%q2l(I)

       profile%L(I) = profile%q2l_b(I)/profile%q2_b(I)
       
       profile%KM(I) = 0.1_RealType*Umol
       profile%KH(I) = 0.1_RealType*Umol
       profile%KQ(I) = 0.1_RealType*Umol
       profile%KH_b(I) = 0.1_RealType*Umol
       profile%KM_b(I) = 0.1_RealType*Umol
    END DO

    !
    ! Set boundary conditions for KQ
    !
    profile%KQ(1) = 0.0_RealType
    profile%KQ(profile%nlevels) = 0.0_RealType

    profile%initialized = .TRUE.

  END SUBROUTINE Initialize_Profile

END MODULE Profile
