!=================================================================
!
! NAME: 
!
!    solar_routines
!
! FUNCTION:
!
!    Routines to calculate the solar flux given the time etc.
!
! TYPE/LANGUAGE: 
!
!    F90 module
!
! DESCRIPTION:  
!
!    Routines to calculate the solar flux for the clear sky case based
!    on time, position, amount of downwelling flux (from a model) as well
!    as calcualting the solar zentih angle and and the clear sky index 
!    (fraction of clear)
!
! ROUTINES TO BE USED EXTERNALLY:
!
!    Solar_Properties  - calculates the solar flux, zenith and and clear
!                        sky index
!
! ROUTINES USED INTERNALLY:
!
!    SolarRadiance     - calculates the direct beam and diffuse radiation
!    solflux           - calculates the direct beam flux
!
! HISTORY:
! 
!   Original version by Jonathan P.D. Mittaz on 9/18/2013
!
!=================================================================
MODULE Solar_Routines

  USE Kinds
  USE Constants

  IMPLICIT NONE

  !
  ! Set only required routines to public
  !
  PRIVATE
  PUBLIC :: Solar_Properties

CONTAINS
  
  !=================================================================
  !
  ! NAME:
  !
  !      Solar_Properties
  !
  ! FUNCTION:
  !
  !      Code to control the computation of the solar zenith angle and clear
  !      sky index from the position and measured downwelling solar radiation.
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Code to control the computation of the solar zenith angle and clear
  !      sky index from the position and measured downwelling solar radiation.
  !
  !      sza and clear sky radiation computed through code based on 
  !      SolarRadiance.m
  !
  !      Note that this code currently requires internal specification of 
  !      atmospheric absorption properties.  At later date these could be 
  !      passed in as arguments.
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !     SolarRadiance
  !
  ! CALLING SEQUENCE:
  !
  !     CALL Solar_Properties( julian_date,latitude,longitude,rs,&
  !                            solar_zenith_angle,clear_sky_index,sw )
  !
  ! INPUTS:
  ! 
  !     julian_date           - Input Julian date to be used for calculation
  !     latitude              - requested latitude
  !     longitude             - requested longitude
  !     downwelling_radiance  - measured downwelling radiation
  !
  ! OUTPUTS:
  !
  !     solar_zenith_angle         - solar zenith angle
  !     clear_sky_index            - clear sky index
  !     total_solar_flux_clear_sky - solar flux for clear sky case
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 7/16/2013
  !      
  !
  SUBROUTINE Solar_Properties(julian_date,latitude,longitude,&
       downwelling_radiance,solar_zenith_angle,clear_sky_index,&
       total_solar_flux_clear_sky)

    REAL(Dble), INTENT(IN) :: julian_date
    REAL(RealType), INTENT(IN) :: latitude
    REAL(RealType), INTENT(IN) :: longitude
    REAL(RealType), INTENT(IN) :: downwelling_radiance
    REAL(RealType), INTENT(OUT) :: solar_zenith_angle
    REAL(RealType), INTENT(OUT) :: clear_sky_index
    REAL(RealType), INTENT(OUT) :: total_solar_flux_clear_sky
    
    ! Local variables
    INTEGER(Int4) :: int_jul_day

    REAL(RealType) :: aerosol_opt_depth_band1
    REAL(RealType) :: aerosol_opt_depth_band2
    REAL(RealType) :: ozone_thickness
    REAL(RealType) :: water_vapor
    REAL(RealType) :: pressure
    REAL(RealType) :: jul_day
    REAL(RealType) :: hour_of_day
    REAL(RealType) :: solar_azimuth
    REAL(RealType) :: direct_beam_irradiance
    REAL(RealType) :: sky

    !
    ! Specify atmospheric absorption properties
    !
    ! At a future date can be obtained from a model
    !
    aerosol_opt_depth_band1 = 0.01_RealType  ! aerosol optical depth, band 1
    aerosol_opt_depth_band2 = 0.01_RealType  ! aerosol optical depth, band 2
    ozone_thickness = 0.2_RealType           ! ozone thickness
    water_vapor = 3.0_RealType               ! g/cm2
    pressure = 1010._RealType                ! pressure hPa

    jul_day = REAL(julian_date)

    !
    ! Original F90 code had INT(jul_day + 30./60./24.) which makes no sense
    ! given below the comment says to preseive full frational hour...
    !
    int_jul_day = INT(jul_day)      ! integer julian day

    !
    ! Comment in original code - kept for consistency
    !
    ! tutc=(jd+30./60./24.-float(jdx))*24.0  !0:1:23;%hour of the day
    ! Extra addition of 30 minutes above only appropriate if were rounding
    ! to integer hour.  Here we preserve full fractional hour.
    !
    ! 0:1:23;%hour of the day
    hour_of_day = (jul_day-float(int_jul_day))*24.0_RealType  

    !
    ! Compute the solar radiance
    !
    call SolarRadiance(latitude,longitude,int_jul_day,hour_of_day,&
         water_vapor,pressure,aerosol_opt_depth_band1,aerosol_opt_depth_band2,&
         ozone_thickness,total_solar_flux_clear_sky,solar_zenith_angle,&
         solar_azimuth,direct_beam_irradiance,sky)

    !
    ! Compute cloud index defined as (1-Ed(measured)/Ed(clear sky))
    !
    IF( total_solar_flux_clear_sky .gt. 0 .and. &
         total_solar_flux_clear_sky .gt. downwelling_radiance )THEN
       clear_sky_index = 1.0_RealType - &
            (downwelling_radiance/total_solar_flux_clear_sky)
    ELSE
       clear_sky_index = 0.0_RealType
    ENDIF
    
  END SUBROUTINE Solar_Properties

  !=================================================================
  !
  ! NAME:
  !
  !      SolarRadiance
  !
  ! FUNCTION:
  !
  !      Code to calculate the solar radiance
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Uses function "solflux" to compute the direct beam and diffuse 
  !      radiation.
  !
  !      Assumptions:
  !          transmittance from aerosols Ta = 0;
  !          sfc pressure = 1023 hPa
  !          ozone = 0.3 (see solflux.m comments);
  !
  !      The ephem function came from P. Minnett
  !      Reynolds, 971029
  !
  !      Note that an azimuth correction may be necessary.  Need to call
  !      routine a second time with a short increment in time.  See 
  !      original matlab code for details
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  !     solflux
  !
  ! CALLING SEQUENCE:
  !
  !     CALL SolarRadiance(latitude,longitude,julian_day,hour_in_day,&
  !                        water_vapour,pressure,aerosol_opt_depth_band1,&
  !                        aerosol_opt_depth_band1,ozone_thickness,&
  !                        total_solar_flux,solar_zenith_angle,solar_azimuth,&
  !                        direct_beam_irradiance,sky)
  !
  ! INPUTS:
  ! 
  !     latitude                - requested latitude
  !     longitude               - requested longitude
  !     julian_day              - integer julian day
  !     hour_in_day             - hour in day
  !     water_vapour            - Water vapor (g/mc2)
  !     pressure                - pressure (hPa)
  !     aerosol_opt_depth_band1 - aerosol optical depth for band 1
  !     aerosol_opt_depth_band2 - aerosol optical depth for band 2
  !     ozone_thickness         - ozone thickness
  !
  ! OUTPUTS:
  !
  !     total_solar_flux        - total solar flux through a horizontal
  !                               surface
  !     solar_zenith_angle      - zenith declination (deg)
  !     solar_azimuth           - solar azimuth (deg)
  !     direct_beam_irradiance  - direct beam irradiance on a surface normal 
  !                               to the beam
  !     sky                     - diffuse component on a horizontal surface
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 7/16/2013
  !      
  !
  SUBROUTINE SolarRadiance(latitude,longitude,julian_day,hour_in_day,&
       water_vapor,pressure,aerosol_opt_depth_band1,aerosol_opt_depth_band2,&
       ozone_thickness,total_solar_flux,solar_zenith_angle,solar_azimuth,&
       direct_beam_irradiance,sky)

    REAL(RealType), INTENT(IN) :: latitude
    REAL(RealType), INTENT(IN) :: longitude
    INTEGER(Int4), INTENT(IN) :: julian_day
    REAL(RealType), INTENT(IN) :: hour_in_day
    REAL(RealType), INTENT(IN) :: water_vapor
    REAL(RealType), INTENT(IN) :: pressure
    REAL(RealType), INTENT(IN) :: aerosol_opt_depth_band1
    REAL(RealType), INTENT(IN) :: aerosol_opt_depth_band2
    REAL(RealType), INTENT(IN) :: ozone_thickness
    REAL(RealType), INTENT(OUT) :: total_solar_flux
    REAL(RealType), INTENT(OUT) :: solar_zenith_angle
    REAL(RealType), INTENT(OUT) :: solar_azimuth
    REAL(RealType), INTENT(OUT) :: direct_beam_irradiance
    REAL(RealType), INTENT(OUT) :: sky

    ! Local variables
    INTEGER(Int4) :: itempterm

    REAL(RealType) :: rlat
    REAL(RealType) :: rlon
    REAL(RealType) :: colatr

    REAL(RealType) :: an
    REAL(RealType) :: al
    REAL(RealType) :: mean_anomoly
    REAL(RealType) :: r
    REAL(RealType) :: s
    REAL(RealType) :: mean_anomoly_rad
    REAL(RealType) :: alambda
    REAL(RealType) :: rlambda
    REAL(RealType) :: epsilon
    REAL(RealType) :: te
    REAL(RealType) :: alpha
    REAL(RealType) :: sin_dec
    REAL(RealType) :: solar_dec
    REAL(RealType) :: cos_dec
    REAL(RealType) :: eq_time
    REAL(RealType) :: s_hour
    REAL(RealType) :: hour_ang
    REAL(RealType) :: cos_h
    REAL(RealType) :: c_solar_z
    REAL(RealType) :: solar_zen
    REAL(RealType) :: solar_az

    ! Convert coordinates to radians
    rlat = latitude*d2r
    rlon = longitude*d2r
    colatr = (90.0_RealType - latitude) * d2r

    !
    ! time check, if we are making a computation at a single time,
    ! we must do it for that time plus a small increment so,
    ! in the end we can adjust the azimuth
    !
    an = FLOAT(julian_day) - 0.5_RealType + hour_in_day/24.0_RealType

    al = 280.460_RealType + 0.9856474_RealType*an
    itempterm = int(al/360.0_RealType)
    al = al - itempterm*360.0_RealType

    IF( al .lt. 0.0_RealType )THEN
       al = al + 360.0_RealType
    ENDIF

    !
    ! CALCULATE MEAN ANOMOLY
    !
    mean_anomoly = 357.528_RealType + 0.9856003_RealType*an
    itempterm = int(mean_anomoly/360.0_RealType)
    mean_anomoly = mean_anomoly - itempterm*360.0_RealType

    IF( mean_anomoly .lt. 0.0_RealType )THEN
       mean_anomoly = mean_anomoly + 360.0_RealType
    ENDIF

    mean_anomoly_rad = mean_anomoly * d2r

    !
    ! SUN EARTH DISTANCE AND NORMAL INSOLATION
    !
    r = 1.00014_RealType - 0.01671_RealType*cos(mean_anomoly_rad) - &
         0.00014_RealType*cos(2*mean_anomoly_rad)
    s = solar_constant/(r*r)

    !
    ! CALCULATE ECLIPTIC LONGITUDE
    !
    alambda = al + 1.915_RealType*sin(mean_anomoly_rad) + &
         0.020_RealType*sin(2*mean_anomoly_rad)
    rlambda = alambda * d2r

    !
    ! ECLIPTIC OBLIQUITY
    !
    epsilon = 23.439_RealType - 0.0000004_RealType*an

    !
    ! RIGHT ASCENSION
    !
    te = tan(epsilon*0.5_RealType*d2r)
    te = te*te
    alpha = alambda - r2d*te*sin(2*rlambda) + &
         0.5_RealType*r2d*te*te*sin(4*rlambda)

    !
    ! SOLAR DECLINATION FOR THIS DAY
    !
    sin_dec = sin(epsilon*d2r) * sin(rlambda)
    solar_dec = asin(sin_dec)
    cos_dec = cos(solar_dec)

    eq_time = (al - alpha)*4/60._RealType
    s_hour = hour_in_day + eq_time
    hour_ang = pi * (s_hour - 12.0_RealType)/12.0_RealType + rlon
    cos_h = cos(hour_ang)
    
    c_solar_z = sin(rlat)*sin_dec + cos(rlat)*cos_dec*cos_h
    solar_zen = acos(c_solar_z)

    !
    ! ZENITH ANGLE
    !
    solar_zenith_angle = solar_zen * r2d
    ! Correct for E-W

    !
    ! AZIMUTH ANGLE
    !
    solar_az = (cos(pi*0.5_RealType - solar_dec) - cos(colatr)*c_solar_z) / &
         (sin(colatr)*sin(solar_zen))
    !
    ! Check against rounding errors - has happend under Mac OS
    !
    IF( solar_az .lt. -1.0_RealType )THEN
       solar_az = -1.0_RealType
    ELSE IF( solar_az .gt. 1.0_RealType )THEN
       solar_az = 1.0_RealType
    ENDIF

    solar_azimuth = acos(solar_az) * r2d

    !
    ! Note that an azimuth correction may be necessary.  Need to call
    ! routine a second time with a short increment in time.  See 
    ! original matlab code for details
    !

    ! AZIMUTH CORRECTION -- where zenith angle is increasing

    !
    ! SHORTWAVE IRRADIANCE
    ! the zenith angle = sz (degs),
    ! the integrated water vapor is set in the input parameters (TWP = 5).
    !
    if (water_vapor .lt. 0.0_RealType) then
       direct_beam_irradiance = s
       sky = 0.0_RealType
       total_solar_flux = s*cos(solar_zenith_angle*d2r)
    else
       CALL solflux(solar_zenith_angle,water_vapor,pressure,&
            aerosol_opt_depth_band1,aerosol_opt_depth_band2,ozone_thickness,&
            direct_beam_irradiance,sky)

       total_solar_flux = sky + direct_beam_irradiance*&
            cos(solar_zenith_angle*d2r)
    end if
    
    if ( total_solar_flux .lt. 0.0_RealType ) then
       total_solar_flux = 0.0_RealType
    end if
    
    if ( sky .lt. 0.0_RealType ) then
       sky = 0.0_RealType
    end if
    
  END SUBROUTINE SolarRadiance    

  !=================================================================
  !
  ! NAME:
  !
  !      solflux
  !
  ! FUNCTION:
  !
  !      Code to calculate direct solar irradiance
  !
  ! TYPE/LANGUAGE:
  !
  !      Fortran 90
  !
  ! DESCRIPTION:
  !
  !      Calculate direct solar irradiance using a parameterization technique
  !      from M. Iqbal in "Physical Climatology for Solar and Wind Energy",
  !      1988, World Scientific, pp. 196-242.
  !
  !      reynolds 981029
  !
  !      Note that only direct and diffuse fluxes are returned in this
  !      version for consistency with the calling code.
  !      No default values for inputs in Fortran code
  !
  ! SUBROUTINE/FUNCTION CALLS:
  !
  ! CALLING SEQUENCE:
  !
  !     CALL solflux(zen_angle,water_vapur,pressure,aerosol_opt_depth_band1,
  !                  aerosol_opt_depth_band2,ozone_thickness,normal_sfc_flux,
  !                  diffuse_sfc_flux)

  !
  ! INPUTS:
  ! 
  !     zen_angle               - solar zenith angle
  !     water_vapour            - Water vapor (g/mc2)
  !     pressure                - pressure (hPa)
  !     aerosol_opt_depth_band1 - aerosol optical depth for band 1
  !     aerosol_opt_depth_band2 - aerosol optical depth for band 2
  !     ozone_thickness         - ozone thickness
  !
  ! OUTPUTS:
  !
  !     normal_sfc_flux         - Direct-Normal flux at surface
  !     diffuse_sfc_flux        - Diffuse flux at surface
  !
  ! RETURNS:
  !
  ! HISTORY
  !
  !      Based on code supplied by Gary Wick NOAA/OAR
  !
  !      Original F90 version by Jonathan P.D. Mittaz 7/16/2013
  !      
  !
  SUBROUTINE solflux(zen_angle,water_vapor,pressure,&
       aerosol_opt_depth_band1,aerosol_opt_depth_band2,ozone_thickness,&
       normal_sfc_flux,diffuse_sfc_flux)

    REAL(RealType), INTENT(INOUT) :: zen_angle
    REAL(RealType), INTENT(IN) :: water_vapor
    REAL(RealType), INTENT(IN) :: pressure
    REAL(RealType), INTENT(IN) :: aerosol_opt_depth_band1
    REAL(RealType), INTENT(IN) :: aerosol_opt_depth_band2
    REAL(RealType), INTENT(IN) :: ozone_thickness
    REAL(RealType), INTENT(OUT) :: normal_sfc_flux
    REAL(RealType), INTENT(OUT) :: diffuse_sfc_flux

    ! Local variables
    REAL(RealType) :: zen_angle_rad

    REAL(RealType) :: Trans_rayleigh    ! Transmittance by Rayleigh scattering
    REAL(RealType) :: Trans_ozone       ! Transmittance by Ozone
    REAL(RealType) :: Trans_gas         ! Transmittance by uniformly mixed gases
    REAL(RealType) :: Trans_water_vapor ! Transmittance by water vapour
    REAL(RealType) :: Trans_aerosol     ! Transmittance by aerosol

    REAL(RealType) :: Taa
    REAL(RealType) :: Idr
    REAL(RealType) :: Tas
    REAL(RealType) :: Ida
    REAL(RealType) :: aa
    REAL(RealType) :: Idm
    REAL(RealType) :: aerosol_opt_depth
    REAL(RealType) :: Ma
    REAL(RealType) :: Mr
    REAL(RealType) :: mu
    REAL(RealType) :: U1
    REAL(RealType) :: U3

    ! 
    ! CLIP ALL ZENITH ANGLES TO <= 89.99 DEG.
    !
    IF( zen_angle .gt. 89.99_RealType )THEN
       zen_angle = 89.99_RealType
       mu = cos(zen_angle*d2r)
    ENDIF

    zen_angle_rad = zen_angle*d2r
    mu = cos(zen_angle_rad)

    !
    !  COMPUTE AIR MASS AND OTHER PROPERTIES
    !
    Mr = 1.0_RealType / &
         (mu + 0.15_RealType*(93.885_RealType - zen_angle)**(-1.253))
    Ma = Mr*(pressure/1013.25_RealType)
    U1 = water_vapor*Mr
    U3 = ozone_thickness*Mr
    aerosol_opt_depth = 0.2758_RealType*aerosol_opt_depth_band1 + &
         0.35_RealType*aerosol_opt_depth_band2

    !
    ! First calculate the transmittances due to various species
    !

    !
    ! RAYLEIGH SCATTERING
    ! 
    IF(zen_angle .le. 85.0_RealType)THEN
       Trans_rayleigh = exp(-0.0903_RealType*Ma**0.84_RealType * &
            (1.0_RealType + Ma - Ma**1.01_RealType))
    ELSE
       Trans_rayleigh = exp(-(0.0738_RealType + &
            (zen_angle-85.0_RealType)*0.03_RealType) * Ma**0.84_RealType)
    ENDIF

    ! 
    !  OZONE SCATTERING
    !
    Trans_ozone = 1.0_RealType - &
         (0.1611_RealType*U3*(1 + 139.48_RealType*U3)**(-0.3035_RealType) - &
         0.002715_RealType*U3/(1 + 0.044_RealType*U3 + 0.0003_RealType*U3*U3))

    !
    !  UNIFORM GASES SCATTERING
    !
    Trans_gas = exp(-0.0127_RealType*Ma**0.26)

    !
    !  WATERVAPOR SCATTERING
    !
    Trans_water_vapor = 1.0_RealType - 2.459_RealType*U1 / &
         ((1.0_RealType+79.034_RealType*U1)**0.6828 + 6.385_RealType*U1)

    !
    !  AEROSOL SCATTERING
    !
    Trans_aerosol = exp(-aerosol_opt_depth**0.873 * &
         (1 + aerosol_opt_depth - aerosol_opt_depth**0.7088) * Ma**0.9108)

    !
    !  direct normal flux
    !

    normal_sfc_flux = solar_constant*Trans_rayleigh*Trans_ozone*Trans_gas*&
         Trans_water_vapor*Trans_aerosol

    !
    !  diffuse component
    !
    Taa = 1 - (1 - w0aer)*(1 - Ma + Ma**1.06)*(1.0_RealType - Trans_aerosol)

    Idr = 0.79_RealType*solar_constant*mu*Trans_ozone * &
         Trans_gas*Trans_water_vapor*Taa * &
         (0.5_RealType*(1-Trans_rayleigh))/(1-Ma+Ma**1.02)
	
    Tas = Trans_aerosol/Taa

    Ida = 0.79*solar_constant*mu*Trans_ozone*Trans_gas*Taa * (fc*(1-Tas)) / &
         (1-Ma+Ma**1.02)

    aa = 0.0685 + (1-fc)*(1-Tas)

    Idm = (normal_sfc_flux*mu + Idr + Ida) * (ag*aa/(1-ag*aa))

    diffuse_sfc_flux = Idr + Ida + Idm

    !
    !  we get negative diffuse components??
    !  This is a quick fix
    !
    IF(diffuse_sfc_flux .lt. 0.0_RealType)THEN
       diffuse_sfc_flux = 0.0_RealType
    ENDIF

  END SUBROUTINE solflux

END MODULE Solar_Routines
